from __future__ import absolute_import  # noreorder

import difflib
import logging
import warnings

from pyaedt import is_ironpython
from pyaedt.edb_core.general import convert_py_list_to_net_list
from pyaedt.generic.clr_module import _clr
from pyaedt.generic.general_methods import pyaedt_function_handler

logger = logging.getLogger(__name__)


class Material(object):
    """Manages EDB methods for material property management."""

    def __init__(self, pclass, edb_material_def):
        self._pclass = pclass
        self._name = edb_material_def.GetName()
        self._edb_material_def = edb_material_def
        self._conductivity = 0.0
        self._loss_tangent = 0.0
        self._magnetic_loss_tangent = 0.0
        self._mass_density = 0.0
        self._permittivity = 0.0
        self._permeability = 0.0
        self._poisson_ratio = 0.0
        self._specific_heat = 0.0
        self._thermal_conductivity = 0.0
        self._youngs_modulus = 0.0
        self._thermal_expansion_coefficient = 0.0
        self._dc_conductivity = 0.0
        self._dc_permittivity = 0.0
        self._dielectric_model_frequency = 0.0
        self._loss_tangent_at_frequency = 0.0
        self._permittivity_at_frequency = 0.0

    def _edb_value(self, value):
        return self._pclass._edb_value(value)

    @property
    def name(self):
        """Retrieve material name."""
        return self._name

    @property
    def _db(self):
        return self._pclass._db

    @property
    def _edb(self):
        return self._pclass._edb

    @pyaedt_function_handler()
    def _get_property(self, property_name):
        if is_ironpython:  # pragma: no cover
            property_box = _clr.StrongBox[float]()
            self._edb_material_def.GetProperty(property_name, property_box)
            return float(property_box)
        else:
            _, property_box = self._edb_material_def.GetProperty(property_name)
            return property_box.ToDouble()

    @property
    def conductivity(self):
        material_id = self._edb.Definition.MaterialPropertyId.Conductivity
        self._conductivity = self._get_property(material_id)
        return self._conductivity

    @conductivity.setter
    def conductivity(self, value):
        """Retrieve material conductivity."""
        material_id = self._edb.Definition.MaterialPropertyId.Conductivity
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._conductivity = value

    @property
    def permittivity(self):
        """Retrieve material permittivity."""
        material_id = self._edb.Definition.MaterialPropertyId.Permittivity
        self._permittivity = self._get_property(material_id)
        return self._permittivity

    @permittivity.setter
    def permittivity(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.Permittivity
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._permittivity = value

    @property
    def permeability(self):
        """Retrieve material permeability."""
        material_id = self._edb.Definition.MaterialPropertyId.Permeability
        self._permeability = self._get_property(material_id)
        return self._permeability

    @permeability.setter
    def permeability(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.Permeability
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._permeability = value

    @property
    def loss_tangent(self):
        """Retrieve material loss tangent."""
        material_id = self._edb.Definition.MaterialPropertyId.DielectricLossTangent
        self._loss_tangent = self._get_property(material_id)
        return self._loss_tangent

    @loss_tangent.setter
    def loss_tangent(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.DielectricLossTangent
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._loss_tangent = value

    @property
    def dc_conductivity(self):
        """"""
        if self._edb_material_def.GetDielectricMaterialModel():
            return self._edb_material_def.GetDielectricMaterialModel().GetDCConductivity()

    @dc_conductivity.setter
    def dc_conductivity(self, value):
        if self._edb_material_def.GetDielectricMaterialModel():
            self._edb_material_def.GetDielectricMaterialModel().SetDCConductivity(value)

    @property
    def dc_permittivity(self):
        """"""
        if self._edb_material_def.GetDielectricMaterialModel():
            return self._edb_material_def.GetDielectricMaterialModel().GetDCRelativePermitivity()

    @dc_permittivity.setter
    def dc_permittivity(self, value):
        if self._edb_material_def.GetDielectricMaterialModel():
            self._edb_material_def.GetDielectricMaterialModel().SetDCRelativePermitivity(value)

    @property
    def dielectric_model_frequency(self):
        """

        Returns
        -------
        Frequency in GHz
        """
        if self._edb_material_def.GetDielectricMaterialModel():
            return self._edb_material_def.GetDielectricMaterialModel().GetFrequency()

    @dielectric_model_frequency.setter
    def dielectric_model_frequency(self, value):
        if self._edb_material_def.GetDielectricMaterialModel():
            self._edb_material_def.GetDielectricMaterialModel().SetFrequency(value)

    @property
    def loss_tangent_at_frequency(self):
        if self._edb_material_def.GetDielectricMaterialModel():
            return self._edb_material_def.GetDielectricMaterialModel().GetLossTangentAtFrequency()

    @loss_tangent_at_frequency.setter
    def loss_tangent_at_frequency(self, value):
        if self._edb_material_def.GetDielectricMaterialModel():
            self._edb_material_def.GetDielectricMaterialModel().SetLossTangentAtFrequency(self._edb_value(value))

    @property
    def permittivity_at_frequency(self):
        if self._edb_material_def.GetDielectricMaterialModel():
            return self._edb_material_def.GetDielectricMaterialModel().GetRelativePermitivityAtFrequency()

    @permittivity_at_frequency.setter
    def permittivity_at_frequency(self, value):
        if self._edb_material_def.GetDielectricMaterialModel():
            self._edb_material_def.GetDielectricMaterialModel().SetRelativePermitivityAtFrequency(value)

    @property
    def magnetic_loss_tangent(self):
        """Retrieve material magnetic loss tangent."""
        material_id = self._edb.Definition.MaterialPropertyId.MagneticLossTangent
        self._magnetic_loss_tangent = self._get_property(material_id)
        return self._magnetic_loss_tangent

    @magnetic_loss_tangent.setter
    def magnetic_loss_tangent(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.MagneticLossTangent
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._magnetic_loss_tangent = value

    @property
    def thermal_conductivity(self):
        """Retrieve material thermal conductivity."""
        material_id = self._edb.Definition.MaterialPropertyId.ThermalConductivity
        self._thermal_conductivity = self._get_property(material_id)
        return self._thermal_conductivity

    @thermal_conductivity.setter
    def thermal_conductivity(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.ThermalConductivity
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._thermal_conductivity = value

    @property
    def mass_density(self):
        """Retrieve material mass density."""
        material_id = self._edb.Definition.MaterialPropertyId.MassDensity
        self._mass_density = self._get_property(material_id)
        return self._mass_density

    @mass_density.setter
    def mass_density(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.MassDensity
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._mass_density = value

    @property
    def youngs_modulus(self):
        """Retrieve material Young's Modulus."""
        material_id = self._edb.Definition.MaterialPropertyId.YoungsModulus
        self._youngs_modulus = self._get_property(material_id)
        return self._youngs_modulus

    @youngs_modulus.setter
    def youngs_modulus(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.YoungsModulus
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._youngs_modulus = value

    @property
    def specific_heat(self):
        """Retrieve material Specific Heat."""
        material_id = self._edb.Definition.MaterialPropertyId.SpecificHeat
        self._specific_heat = self._get_property(material_id)
        return self._specific_heat

    @specific_heat.setter
    def specific_heat(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.SpecificHeat
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._specific_heat = value

    @property
    def poisson_ratio(self):
        """Retrieve material Poisson Ratio."""
        material_id = self._edb.Definition.MaterialPropertyId.PoissonsRatio
        self._poisson_ratio = self._get_property(material_id)
        return self._poisson_ratio

    @poisson_ratio.setter
    def poisson_ratio(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.PoissonsRatio
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._poisson_ratio = value

    @property
    def thermal_expansion_coefficient(self):
        """Retrieve material Thermal Coefficient.."""
        material_id = self._edb.Definition.MaterialPropertyId.ThermalExpansionCoefficient
        self._thermal_expansion_coefficient = self._get_property(material_id)
        return self._thermal_expansion_coefficient

    @thermal_expansion_coefficient.setter
    def thermal_expansion_coefficient(self, value):
        material_id = self._edb.Definition.MaterialPropertyId.ThermalExpansionCoefficient
        self._edb_material_def.SetProperty(material_id, self._edb_value(value))
        self._thermal_expansion_coefficient = value

    @pyaedt_function_handler()
    def _json_format(self):
        out_dict = {}
        if self.permittivity == 0:  # pragma no cover
            self.permittivity = 1.0
        if self.permeability == 0:  # pragma no cover
            self.permeability = 1.0
        self._name = self.name
        self._conductivity = self.conductivity
        self._loss_tangent = self.loss_tangent
        self._magnetic_loss_tangent = self.magnetic_loss_tangent
        self._mass_density = self.mass_density
        self._permittivity = self.permittivity
        self._permeability = self.permeability
        self._poisson_ratio = self.poisson_ratio
        self._specific_heat = self.specific_heat
        self._thermal_conductivity = self.thermal_conductivity
        self._youngs_modulus = self.youngs_modulus
        self._thermal_expansion_coefficient = self.thermal_expansion_coefficient
        self._dc_conductivity = self.dc_conductivity
        self._dc_permittivity = self.dc_permittivity
        self._dielectric_model_frequency = self.dielectric_model_frequency
        self._loss_tangent_at_frequency = self.loss_tangent_at_frequency
        self._permittivity_at_frequency = self.permittivity_at_frequency
        for k, v in self.__dict__.items():
            if not k == "_pclass" and not k == "_edb_material_def":
                out_dict[k[1:]] = v
        return out_dict

    @pyaedt_function_handler()
    def _load(self, input_dict):
        if input_dict:
            self.conductivity = input_dict["conductivity"]
            self.loss_tangent = input_dict["loss_tangent"]
            self.magnetic_loss_tangent = input_dict["magnetic_loss_tangent"]
            self.mass_density = input_dict["mass_density"]
            self.permittivity = input_dict["permittivity"]
            self.permeability = input_dict["permeability"]
            self.poisson_ratio = input_dict["poisson_ratio"]
            self.specific_heat = input_dict["specific_heat"]
            self.thermal_conductivity = input_dict["thermal_conductivity"]
            self.youngs_modulus = input_dict["youngs_modulus"]
            self.thermal_expansion_coefficient = input_dict["thermal_expansion_coefficient"]
            if input_dict["dielectric_model_frequency"] is not None:
                if self._edb_material_def.GetDielectricMaterialModel():
                    model = self._edb_material_def.GetDielectricMaterialModel()
                    self.dielectric_model_frequency = input_dict["dielectric_model_frequency"]
                    self.loss_tangent_at_frequency = input_dict["loss_tangent_at_frequency"]
                    self.permittivity_at_frequency = input_dict["permittivity_at_frequency"]
                    if input_dict["dc_permittivity"] is not None:
                        model.SetUseDCRelativePermitivity(True)
                        self.dc_permittivity = input_dict["dc_permittivity"]
                    self.dc_conductivity = input_dict["dc_conductivity"]
                else:
                    if not self._pclass.add_djordjevicsarkar_material(
                        input_dict["name"],
                        input_dict["permittivity_at_frequency"],
                        input_dict["loss_tangent_at_frequency"],
                        input_dict["dielectric_model_frequency"],
                        input_dict["dc_permittivity"],
                        input_dict["dc_conductivity"],
                    ):
                        self._pclass._pedb.logger.warning(
                            'Cannot set DS model for material "{}". Check for realistic '
                            "values that define DS Model".format(input_dict["name"])
                        )
            else:
                # To unset DS model if already assigned to the material in database
                if self._edb_material_def.GetDielectricMaterialModel():
                    self._edb_material_def.SetDielectricMaterialModel(self._edb_value(None))


class Materials(object):
    """Manages EDB methods for material management accessible from `Edb.materials` property."""

    def __getitem__(self, item):
        return self.materials[item]

    def __init__(self, pedb):
        self._pedb = pedb
        if not self.materials:
            self.add_material("air")
            self.add_material("copper", 1, 0.999991, 5.8e7, 0, 0)
            self.add_material("fr4_epoxy", 4.4, 1, 0, 0.02, 0)
            self.add_material("solder_mask", 3.1, 1, 0, 0.035, 0)

    @pyaedt_function_handler()
    def _edb_value(self, value):
        return self._pedb.edb_value(value)

    @property
    def _edb(self):
        return self._pedb.edb

    @property
    def _db(self):
        return self._pedb.db

    @property
    def materials(self):
        """Retrieve dictionary of material from material library."""
        return {obj.GetName(): Material(self, obj) for obj in list(self._db.MaterialDefs)}

    @pyaedt_function_handler
    def add_material(
        self,
        name="air",
        permittivity=1.0006,
        permeability=1.0000004,
        conductivity=0,
        dielectric_loss_tangent=0,
        magnetic_loss_tangent=0,
    ):
        """Add a new material.

        Parameters
        ----------
        name : str, optional
            Material Name. The default is ``"air"``.
        permittivity : float, str, optional
            Material permittivity. The default is ``1.0006``.
        permeability : float, str, optional
            Material permeability. The default is ``1.0000004``.
        conductivity : float, str, optional
            Material conductivity. The default is ``0``.
        dielectric_loss_tangent : float, str, optional
            Material dielectric loss tangent. The default is ``0``.
        magnetic_loss_tangent : float, str, optional
            Material magnetic loss tangent. The default is ``0``.

        Returns
        -------
        :class:`pyaedt.edb_core.materials.Material`
        """
        if not name in self.materials:
            self._edb.Definition.MaterialDef.Create(self._db, name)
            new_material = self.materials[name]
            new_material.permittivity = permittivity
            new_material.permeability = permeability
            new_material.conductivity = conductivity
            new_material.loss_tangent = dielectric_loss_tangent
            new_material.magnetic_loss_tangent = magnetic_loss_tangent
            return new_material
        else:  # pragma: no cover
            warnings.warn("Material {} already exists in material library.".format(name))
            return False

    @pyaedt_function_handler()
    def add_conductor_material(self, name, conductivity):
        """Add a new conductor material in library.

        Parameters
        ----------
        name : str
            Name of the new material.
        conductivity : float
            Conductivity of the new material.

        Returns
        -------
        :class:`pyaedt.edb_core.materials.Material`

        """
        if not name in self.materials:
            self._edb.Definition.MaterialDef.Create(self._db, name)
            new_material = self.materials[name]
            new_material.conductivity = conductivity
            new_material.permittivity = 1
            new_material.permeability = 1
            return new_material
        else:  # pragma: no cover
            warnings.warn("Material {} already exists in material library.".format(name))
            return False

    @pyaedt_function_handler()
    def add_dielectric_material(self, name, permittivity, loss_tangent, permeability=1):
        """Add a new dielectric material in library.

        Parameters
        ----------
        name : str
            Name of the new material.
        permittivity : float
            Permittivity of the new material.
        loss_tangent : float
            Loss tangent of the new material.
        permeability: float
            Permeability of the new material.

        Returns
        -------
        :class:`pyaedt.edb_core.materials.Material`
        """
        if not name in self.materials:
            self._edb.Definition.MaterialDef.Create(self._db, name)
            new_material = self.materials[name]
            new_material.permittivity = permittivity
            new_material.loss_tangent = loss_tangent
            new_material.permeability = permeability
            return new_material
        else:
            warnings.warn("Material {} already exists in material library.".format(name))
            return False

    @pyaedt_function_handler()
    def get_djordjevicsarkar_model(self, material_name=None):
        """Djordjevic-Sarkar model if present.

        Parameters
        ----------
        material_name : str

        Returns
        -------

        """
        material = self.materials[material_name]
        if material:
            return material.GetDielectricMaterialModel()

    @pyaedt_function_handler()
    def add_djordjevicsarkar_material(
        self, name, permittivity, loss_tangent, test_frequency, dc_permittivity=None, dc_conductivity=None
    ):
        """Create a Djordjevic_Sarkar dielectric.

        Parameters
        ----------
        name : str
            Name of the dielectric.
        permittivity : float
            Relative permittivity of the dielectric.
        loss_tangent : float
            Loss tangent for the material.
        test_frequency : float
            Test frequency in GHz for the dielectric.
        dc_permittivity : float, optional
            DC Relative permittivity of the dielectric.
        dc_conductivity : float, optional
            DC Conductivity of the dielectric.

        Returns
        -------
        :class:`pyaedt.edb_core.materials.Material`
            Material definition.
        """
        material_def = self._edb.Definition.DjordjecvicSarkarModel()
        material_def.SetFrequency(test_frequency)
        material_def.SetLossTangentAtFrequency(self._edb_value(loss_tangent))
        material_def.SetRelativePermitivityAtFrequency(permittivity)
        if dc_conductivity is not None:
            material_def.SetDCConductivity(dc_conductivity)
        if dc_permittivity is not None:
            material_def.SetUseDCRelativePermitivity(True)
            material_def.SetDCRelativePermitivity(dc_permittivity)
        return self._add_dielectric_material_model(name, material_def)

    @pyaedt_function_handler()
    def add_debye_material(
        self,
        name,
        permittivity_low,
        permittivity_high,
        loss_tangent_low,
        loss_tangent_high,
        lower_freqency,
        higher_frequency,
    ):
        """Create a dielectric with the Debye model.

        Parameters
        ----------
        name : str
            Name of the dielectric.
        permittivity_low : float
            Relative permittivity of the dielectric at the frequency specified
            for ``lower_frequency``.
        permittivity_high : float
            Relative permittivity of the dielectric at the frequency specified
            for ``higher_frequency``.
        loss_tangent_low : float
            Loss tangent for the material at the frequency specified
            for ``lower_frequency``.
        loss_tangent_high : float
            Loss tangent for the material at the frequency specified
            for ``higher_frequency``.
        lower_freqency : float
            Value for the lower frequency.
        higher_frequency : float
            Value for the higher frequency.

        Returns
        -------
        :class:`pyaedt.edb_core.materials.Material`
            Material definition.
        """
        material_def = self._edb.Definition.DebyeModel()
        material_def.SetFrequencyRange(lower_freqency, higher_frequency)
        material_def.SetLossTangentAtHighLowFrequency(loss_tangent_low, loss_tangent_high)
        material_def.SetRelativePermitivityAtHighLowFrequency(
            self._edb_value(permittivity_low), self._edb_value(permittivity_high)
        )
        return self._add_dielectric_material_model(name, material_def)

    @pyaedt_function_handler()
    def add_multipole_debye_material(
        self,
        name,
        frequencies,
        permittivities,
        loss_tangents,
    ):
        """Create a dielectric with the Multipole Debye model.

        Parameters
        ----------
        name : str
            Name of the dielectric.
        frequencies : list
            Frequencies in GHz.
        permittivities : list
            Relative permittivities at each frequency.
        loss_tangents : list
            Loss tangents at each frequency.

        Returns
        -------
        :class:`pyaedt.edb_core.materials.Material`
            Material definition.

        Examples
        --------
        >>> from pyaedt import Edb
        >>> edb = Edb()
        >>> freq = [0, 2, 3, 4, 5, 6]
        >>> rel_perm = [1e9, 1.1e9, 1.2e9, 1.3e9, 1.5e9, 1.6e9]
        >>> loss_tan = [0.025, 0.026, 0.027, 0.028, 0.029, 0.030]
        >>> diel = edb.materials.add_multipole_debye_material("My_MP_Debye", freq, rel_perm, loss_tan)
        """
        frequencies = [float(i) for i in frequencies]
        permittivities = [float(i) for i in permittivities]
        loss_tangents = [float(i) for i in loss_tangents]
        material_def = self._edb.Definition.MultipoleDebyeModel()
        material_def.SetParameters(
            convert_py_list_to_net_list(frequencies),
            convert_py_list_to_net_list(permittivities),
            convert_py_list_to_net_list(loss_tangents),
        )
        return self._add_dielectric_material_model(name, material_def)

    @pyaedt_function_handler()
    def _add_dielectric_material_model(self, name, material_model):
        if self._edb.Definition.MaterialDef.FindByName(self._db, name).IsNull():
            self._edb.Definition.MaterialDef.Create(self._db, name)
        material_def = self._edb.Definition.MaterialDef.FindByName(self._db, name)
        succeeded = material_def.SetDielectricMaterialModel(material_model)
        if succeeded:
            return material_def
        return False

    @pyaedt_function_handler()
    def duplicate(self, material_name, new_material_name):
        """Duplicate a material from the database.


        Parameters
        ----------
        material_name : str
            Name of the existing material.
        new_material_name : str
            Name of the new duplicated material.

        Returns
        -------
        EDB material : class: 'Ansys.Ansoft.Edb.Definition.MaterialDef'


        Examples
        --------

        >>> from pyaedt import Edb
        >>> edb_app = Edb()
        >>> my_material = edb_app.materials.duplicate("copper", "my_new_copper")

        """
        material_list = {i.lower(): i for i in list(self.materials.keys())}
        if material_name.lower() in material_list and new_material_name not in self.materials:
            material_name = material_list[material_name.lower()]
            permittivity = self._edb_value(self.materials[material_name].permittivity)
            permeability = self._edb_value(self.materials[material_name].permeability)
            conductivity = self._edb_value(self.materials[material_name].conductivity)
            dielectric_loss_tangent = self._edb_value(self.materials[material_name].loss_tangent)
            magnetic_loss_tangent = self._edb_value(self.materials[material_name].magnetic_loss_tangent)
            thermal_conductivity = self._edb_value(self.materials[material_name].thermal_conductivity)
            thermal_expansion_coefficient = self._edb_value(self.materials[material_name].thermal_expansion_coefficient)
            youngs_modulus = self._edb_value(self.materials[material_name].youngs_modulus)
            poisson_ratio = self._edb_value(self.materials[material_name].poisson_ratio)
            mass_density = self._edb_value(self.materials[material_name].mass_density)
            material_model = self.materials[material_name]._edb_material_def.GetDielectricMaterialModel()
            edb_material = self._edb.Definition.MaterialDef.Create(self._db, new_material_name)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.Permittivity, permittivity)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.Permeability, permeability)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.Conductivity, conductivity)
            edb_material.SetProperty(
                self._edb.Definition.MaterialPropertyId.DielectricLossTangent, dielectric_loss_tangent
            )
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.ThermalConductivity, thermal_conductivity)
            edb_material.SetProperty(
                self._edb.Definition.MaterialPropertyId.ThermalExpansionCoefficient, thermal_expansion_coefficient
            )
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.MassDensity, mass_density)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.YoungsModulus, youngs_modulus)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.PoissonsRatio, poisson_ratio)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.MagneticLossTangent, magnetic_loss_tangent)
            edb_material.SetProperty(self._edb.Definition.MaterialPropertyId.MagneticLossTangent, magnetic_loss_tangent)
            edb_material.SetDielectricMaterialModel(material_model)

            return edb_material

    @pyaedt_function_handler()
    def _load_materials(self, material=None):
        if self.materials:
            mat_keys = [i.lower() for i in self.materials.keys()]
            mat_keys_case = [i for i in self.materials.keys()]
        else:
            mat_keys = []
            mat_keys_case = []

        if not material:
            return
        if material["name"].lower() not in mat_keys:
            if material["conductivity"] > 1e4:
                self.add_conductor_material(material["name"], material["conductivity"])
            else:
                self.add_dielectric_material(material["name"], material["permittivity"], material["loss_tangent"])
            self.materials[material["name"]]._load(material)
        else:
            self.materials[mat_keys_case[mat_keys.index(material["name"].lower())]]._load(material)

    @pyaedt_function_handler
    def material_name_to_id(self, property_name):
        """Convert a material property name to a material property ID.

        Parameters
        ----------
        property_name : str
            Name of the material property.

        Returns
        -------
        ID of the material property.
        """
        props = {
            "Permittivity": self._edb.Definition.MaterialPropertyId.Permittivity,
            "Permeability": self._edb.Definition.MaterialPropertyId.Permeability,
            "Conductivity": self._edb.Definition.MaterialPropertyId.Conductivity,
            "DielectricLossTangent": self._edb.Definition.MaterialPropertyId.DielectricLossTangent,
            "MagneticLossTangent": self._edb.Definition.MaterialPropertyId.MagneticLossTangent,
            "ThermalConductivity": self._edb.Definition.MaterialPropertyId.ThermalConductivity,
            "MassDensity": self._edb.Definition.MaterialPropertyId.MassDensity,
            "SpecificHeat": self._edb.Definition.MaterialPropertyId.SpecificHeat,
            "YoungsModulus": self._edb.Definition.MaterialPropertyId.YoungsModulus,
            "PoissonsRatio": self._edb.Definition.MaterialPropertyId.PoissonsRatio,
            "ThermalExpansionCoefficient": self._edb.Definition.MaterialPropertyId.ThermalExpansionCoefficient,
            "InvalidProperty": self._edb.Definition.MaterialPropertyId.InvalidProperty,
        }

        found_el = difflib.get_close_matches(property_name, list(props.keys()), 1, 0.7)
        if found_el:
            return props[found_el[0]]
        else:
            return self._edb.Definition.MaterialPropertyId.InvalidProperty

    @pyaedt_function_handler()
    def get_property_by_material_name(self, property_name, material_name):
        """Get the property of a material. If it is executed in IronPython,
         you must only use the first element of the returned tuple, which is a float.

        Parameters
        ----------
        material_name : str
            Name of the existing material.
        property_name : str
            Name of the material property.
            ``permittivity``
            ``permeability``
            ``conductivity``
            ``dielectric_loss_tangent``
            ``magnetic_loss_tangent``

        Returns
        -------
        float
            The float value of the property.


        Examples
        --------
        >>> from pyaedt import Edb
        >>> edb_app = Edb()
        >>> returned_tuple = edb_app.core_stackup.get_property_by_material_name("conductivity", "copper")
        >>> edb_value = returned_tuple[0]
        >>> float_value = returned_tuple[1]

        """
        if self._edb.Definition.MaterialDef.FindByName(self._pedb._db, material_name).IsNull():
            self._pedb.logger.error("This material doesn't exists.")
        else:
            original_material = self._edb.Definition.MaterialDef.FindByName(self._pedb._db, material_name)
            if is_ironpython:  # pragma: no cover
                property_box = _clr.StrongBox[float]()
                original_material.GetProperty(self.material_name_to_id(property_name), property_box)
                return float(property_box)
            else:
                _, property_box = original_material.GetProperty(
                    self.material_name_to_id(property_name), self._edb_value(0.0)
                )
                return property_box.ToDouble()
        return False



from pyaedt.edb_core.ipc2581.ecad.cad_data.padstack_hole_def import PadstackHoleDef
from pyaedt.edb_core.ipc2581.ecad.cad_data.padstack_pad_def import PadstackPadDef
from pyaedt.generic.general_methods import ET


class PadstackDef(object):
    """Class describing an IPC2581 padstack definition."""

    def __init__(self):
        self.name = ""
        self.padstack_hole_def = PadstackHoleDef()
        self._padstack_pad_def = []

    @property
    def padstack_pad_def(self):
        return self._padstack_pad_def

    @padstack_pad_def.setter
    def padstack_pad_def(self, value):  # pragma no cover
        if isinstance(value, list):
            if len([pad for pad in value if isinstance(pad, PadstackPadDef)]) == len(value):
                self._padstack_pad_def = value

    def add_padstack_pad_def(self, layer="", pad_use="REGULAR", x="0", y="0", primitive_ref=""):  # pragma no cover
        pad = PadstackPadDef()
        pad.layer_ref = layer
        pad.pad_use = pad_use
        pad.x = x
        pad.y = y
        pad.primitive_ref = primitive_ref
        self.padstack_pad_def.append(pad)

    def write_xml(self, step):  # pragma no cover
        padstack_def = ET.SubElement(step, "PadStackDef")
        padstack_def.set("name", self.name)
        self.padstack_hole_def.write_xml(padstack_def)
        for pad in self.padstack_pad_def:
            pad.write_xml(padstack_def)



from pyaedt.edb_core.ipc2581.content.entry_line import EntryLine
from pyaedt.generic.general_methods import ET


class DictionaryLine(object):
    def __init__(self, content):
        self._dict_lines = {}
        self.units = content.units

    @property
    def dict_lines(self):
        return self._dict_lines

    @dict_lines.setter
    def dict_lines(self, value):  # pragma no cover
        if isinstance(value, EntryLine):
            self._dict_lines = value

    def add_line(self, width=None):  # pragma no cover
        if width:
            line = EntryLine()
            line._line_width = width
            if not "ROUND_{}".format(width) in self._dict_lines:
                self._dict_lines["ROUND_{}".format(width)] = line

    def write_xml(self, content=None):  # pragma no cover
        dict_line = ET.SubElement(content, "DictionaryLineDesc")
        dict_line.set("units", self.units)
        for line in list(self._dict_lines.values()):
            line.write_xml(dict_line)



from pyaedt.edb_core.ipc2581.content.color import Color
from pyaedt.generic.general_methods import ET


class EntryColor(object):
    def __init__(self):
        self.name = ""
        self.color = Color()

    def write_xml(self, color=None):  # pragma no cover
        entry_color = ET.SubElement(color, "EntryColor")
        entry_color.set("id", self.name)
        self.color.write_xml(entry_color)



# -*- coding: utf-8 -*-
"""
This module contains these classes: `Components3DLayout`,`CircuitComponent',
`EdgePrimitive`, `EdgeTypePrimitive`, `FacePrimitive`, `Geometries3DLayout`,
`Nets3DLayout`, `Objec3DLayout`, `Object3d`, `Padstack`, `PDSHole`, `PDSLayer`,
`Pins3DLayout', and `VertexPrimitive`.

This module provides methods and data structures for managing all properties of
objects (points, lines, sheeets, and solids) within the AEDT 3D Modeler.

"""
from __future__ import absolute_import  # noreorder

import os
import re
import warnings

from pyaedt.generic.constants import AEDT_UNITS
from pyaedt.generic.general_methods import _retry_ntimes
from pyaedt.generic.general_methods import _to_boolean
from pyaedt.generic.general_methods import _uname
from pyaedt.generic.general_methods import clamp
from pyaedt.generic.general_methods import is_ironpython
from pyaedt.generic.general_methods import open_file
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.generic.general_methods import rgb_color_codes
from pyaedt.generic.general_methods import settings
from pyaedt.modeler.cad.elements3d import BinaryTreeNode
from pyaedt.modeler.cad.elements3d import EdgePrimitive
from pyaedt.modeler.cad.elements3d import FacePrimitive
from pyaedt.modeler.cad.elements3d import VertexPrimitive


class Object3d(object):
    """Manages object attributes for the AEDT 3D Modeler.

    Parameters
    ----------
    primitives : :class:`pyaedt.modeler.Primitives3D.Primitives3D`
        Inherited parent object.
    name : str

    Examples
    --------
    Basic usage demonstrated with an HFSS design:

    >>> from pyaedt import Hfss
    >>> aedtapp = Hfss()
    >>> prim = aedtapp.modeler

    Create a part, such as box, to return an :class:`pyaedt.modeler.Object3d.Object3d`.

    >>> id = prim.create_box([0, 0, 0], [10, 10, 5], "Mybox", "Copper")
    >>> part = prim[id]
    """

    def __init__(self, primitives, name=None):
        """
        Parameters
        ----------
        primitives : :class:`pyaedt.modeler.Primitives3D.Primitives3D`
            Inherited parent object.
        name : str
        """
        self._id = None
        if name:
            self._m_name = name
        else:
            self._m_name = _uname()
        self._primitives = primitives
        self.flags = ""
        self._part_coordinate_system = "Global"
        self._material_name = None
        self._transparency = None
        self._solve_inside = None
        self._is_updated = False
        self._all_props = None
        self._surface_material = None
        self._color = None
        self._wireframe = None
        self._part_coordinate_system = None
        self._model = None
        self._m_groupName = None
        self._object_type = None
        self._mass = 0.0
        self._volume = 0.0

    @pyaedt_function_handler()
    def _bounding_box_unmodel(self):
        """Bounding box of a part, unmodel/undo method.

        This is done by setting all other objects as unmodel and getting the model bounding box.
        Then an undo operation restore the design.

        Returns
        -------
        list of [list of float]
            List of six ``[x, y, z]`` positions of the bounding box containing
            Xmin, Ymin, Zmin, Xmax, Ymax, and Zmax values.

        """
        objs_to_unmodel = [
            val.name for i, val in self._primitives.objects.items() if val.model and val.name != self.name
        ]
        if objs_to_unmodel:
            vArg1 = ["NAME:Model", "Value:=", False]
            self._primitives._change_geometry_property(vArg1, objs_to_unmodel)
        modeled = True
        if not self.model:
            vArg1 = ["NAME:Model", "Value:=", True]
            self._primitives._change_geometry_property(vArg1, self.name)
            modeled = False
        bounding = self._primitives.get_model_bounding_box()
        if objs_to_unmodel:
            self._odesign.Undo()
        if not modeled:
            self._odesign.Undo()
        if not settings.non_graphical:
            self._primitives._app.odesktop.ClearMessages(
                self._primitives._app.project_name, self._primitives._app.design_name, 1
            )
        return bounding

    @pyaedt_function_handler()
    def _bounding_box_sat(self):
        """Bounding box of a part.

        This is done by exporting a part as a SAT file and reading the bounding box information from the SAT file.
        A list of six 3D position vectors is returned.

        Returns
        -------
        list of [list of float]
            List of six ``[x, y, z]`` positions of the bounding box containing
            Xmin, Ymin, Zmin, Xmax, Ymax, and Zmax values.

        References
        ----------

        >>> oEditor.GetModelBoundingBox

        """
        tmp_path = self._primitives._app.working_directory
        filename = os.path.join(tmp_path, self.name + ".sat")

        self._primitives._app.export_3d_model(self.name, tmp_path, ".sat", [self.name])

        if not os.path.isfile(filename):
            raise Exception("Cannot export the ACIS SAT file for object {}".format(self.name))

        with open_file(filename, "r") as fh:
            temp = fh.read().splitlines()
        all_lines = [s for s in temp if s.startswith("body ")]

        bb = []
        if len(all_lines) == 1:
            line = all_lines[0]
            pattern = r".+ (.+) (.+) (.+) (.+) (.+) (.+) #$"
            m = re.search(pattern, line)
            if m:
                try:
                    for i in range(1, 7):
                        bb.append(float(m.group(i)))
                except:
                    return False
            else:
                return False
        else:
            return False

        try:
            os.remove(filename)
        except:
            self.logger.warning("ERROR: Cannot remove temp file.")
        return bb

    @property
    def bounding_box(self):
        """Bounding box of a part.

        A list of six 3D position vectors is returned.

        Returns
        -------
        list of [list of float]
            List of six ``[x, y, z]`` positions of the bounding box containing
            Xmin, Ymin, Zmin, Xmax, Ymax, and Zmax values.

        References
        ----------

        >>> oEditor.GetModelBoundingBox

        """
        if self.object_type == "Unclassified":
            return [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        if not settings.disable_bounding_box_sat:
            bounding = self._bounding_box_sat()
            if bounding:
                return bounding
            else:
                return self._bounding_box_unmodel()
        else:
            return self._bounding_box_unmodel()

    @property
    def bounding_dimension(self):
        """Retrieve the dimension array of the bounding box.

        Returns
        -------
        list
            List of three float values representing the bounding box dimensions
            in the form ``[dim_x, dim_y, dim_z]``.

        References
        ----------

        >>> oEditor.GetModelBoundingBox
        """
        oBoundingBox = self.bounding_box
        dimensions = []
        dimensions.append(abs(float(oBoundingBox[0]) - float(oBoundingBox[3])))
        dimensions.append(abs(float(oBoundingBox[1]) - float(oBoundingBox[4])))
        dimensions.append(abs(float(oBoundingBox[2]) - float(oBoundingBox[5])))
        return dimensions

    @property
    def _odesign(self):
        """Design."""
        return self._primitives._modeler._app._odesign

    @pyaedt_function_handler()
    def plot(self, show=True):
        """Plot model with PyVista.

        Parameters
        ----------
        show : bool, optional
            Show the plot after generation.  The default value is ``True``.

        Returns
        -------
        :class:`pyaedt.generic.plot.ModelPlotter`
            Model Object.

        Notes
        -----
        Works from AEDT 2021.2 in CPython only. PyVista has to be installed.
        """
        if not is_ironpython and self._primitives._app._aedt_version >= "2021.2":
            return self._primitives._app.post.plot_model_obj(
                objects=[self.name],
                plot_as_separate_objects=True,
                plot_air_objects=True,
                show=show,
            )

    @pyaedt_function_handler()
    def export_image(self, file_path=None):
        """Export the current object to a specified file path.


        .. note::
           Works from AEDT 2021.2 in CPython only. PyVista has to be installed.

        Parameters
        ----------
        file_path : str, optional
            File name with full path. If `None` the exported image will be a ``png`` file that
            will be saved in ``working_directory``.
            To access the ``working_directory`` the use ``app.working_directory`` property.

        Returns
        -------
        str
            File path.
        """
        if not is_ironpython and self._primitives._app._aedt_version >= "2021.2":
            if not file_path:
                file_path = os.path.join(self._primitives._app.working_directory, self.name + ".png")
            model_obj = self._primitives._app.post.plot_model_obj(
                objects=[self.name],
                show=False,
                export_path=file_path,
                plot_as_separate_objects=True,
                clean_files=True,
            )
            if model_obj:
                return model_obj.image_file
        return False

    @property
    def touching_objects(self):
        """Get the objects that touch one of the vertex, edge midpoint or face of the object."""
        if self.object_type == "Unclassified":
            return []
        list_names = []
        for vertex in self.vertices:
            body_names = self._primitives.get_bodynames_from_position(vertex.position)
            a = [i for i in body_names if i != self.name and i not in list_names]
            if a:
                list_names.extend(a)
        for edge in self.edges:
            body_names = self._primitives.get_bodynames_from_position(edge.midpoint)
            a = [i for i in body_names if i != self.name and i not in list_names]
            if a:
                list_names.extend(a)
        for face in self.faces:
            body_names = self._primitives.get_bodynames_from_position(face.center)
            a = [i for i in body_names if i != self.name and i not in list_names]
            if a:
                list_names.extend(a)
        return list_names

    @pyaedt_function_handler()
    def get_touching_faces(self, object_name):
        """Get the objects that touch one of the face center of each face of the object.

        Parameters
        ----------
        object_name : str, :class:`Object3d`
            Object to check.
        Returns
        -------
        list
            list of objects and faces touching."""

        _names = []
        if isinstance(object_name, Object3d):
            object_name = object_name.name
        for face in self.faces:
            body_names = self._primitives.get_bodynames_from_position(face.center)
            if object_name in body_names:
                _names.append(face)
        return _names

    @property
    def faces(self):
        """Information for each face in the given part.

        Returns
        -------
        list of :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.GetFaceIDs

        """
        if self.object_type == "Unclassified":
            return []
        faces = []
        for face in list(self.m_Editor.GetFaceIDs(self.name)):
            face = int(face)
            faces.append(FacePrimitive(self, face))
        return faces

    @property
    def faces_on_bounding_box(self):
        """Return only the face ids of the faces touching the bounding box.

        Returns
        -------
        List of :class:`pyaedt.modeler.Object3d.FacePrimitive`
        """
        f_list = []
        for face in self.faces:
            if face.is_on_bounding():
                f_list.append(face)
        return f_list

    @property
    def face_closest_to_bounding_box(self):
        """Return only the face ids of the face closest to the bounding box.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`
        """
        b = [float(i) for i in list(self.m_Editor.GetModelBoundingBox())]
        f_id = None
        f_val = None
        for face in self.faces:
            c = face.center
            p_dist = min(
                [
                    abs(c[0] - b[0]),
                    abs(c[1] - b[1]),
                    abs(c[2] - b[2]),
                    abs(c[0] - b[3]),
                    abs(c[1] - b[4]),
                    abs(c[2] - b[5]),
                ]
            )

            if f_val and p_dist < f_val or not f_val:
                f_id = face
                f_val = p_dist
        return f_id

    @pyaedt_function_handler()
    def largest_face(self, n=1):
        """Return only the face with the greatest area.

        Returns
        -------
        List of :class:`pyaedt.modeler.Object3d.FacePrimitive`
        """
        f = []
        for face in self.faces:
            f.append((face.area, face))
        f.sort(key=lambda tup: tup[0], reverse=True)
        f_sorted = [x for y, x in f]
        return f_sorted[:n]

    @pyaedt_function_handler()
    def longest_edge(self, n=1):
        """Return only the edge with the greatest length.

        Returns
        -------
        List of :class:`pyaedt.modeler.Object3d.EdgePrimitive`
        """
        e = []
        for edge in self.edges:
            e.append((edge.length, edge))
        e.sort(key=lambda tup: tup[0], reverse=True)
        e_sorted = [x for y, x in e]
        return e_sorted[:n]

    @pyaedt_function_handler()
    def smallest_face(self, n=1):
        """Return only the face with the smallest area.

        Returns
        -------
        List of :class:`pyaedt.modeler.Object3d.FacePrimitive`
        """
        f = []
        for face in self.faces:
            f.append((face.area, face))
        f.sort(key=lambda tup: tup[0])
        f_sorted = [x for y, x in f]
        return f_sorted[:n]

    @pyaedt_function_handler()
    def shortest_edge(self, n=1):
        """Return only the edge with the smallest length.

        Returns
        -------
        List of :class:`pyaedt.modeler.Object3d.EdgePrimitive`
        """
        e = []
        for edge in self.edges:
            e.append((edge.length, edge))
        e.sort(
            key=lambda tup: tup[0],
        )
        e_sorted = [x for y, x in e]
        return e_sorted[:n]

    @property
    def top_face_z(self):
        """Top face in the Z direction of the object.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.center[2]), face) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[-1][1]
        except:
            return None

    @property
    def bottom_face_z(self):
        """Bottom face in the Z direction of the object.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.center[2]), face) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[0][1]
        except:
            return None

    @property
    def top_face_x(self):
        """Top face in the X direction of the object.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.center[0]), face) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[-1][1]
        except:
            return None

    @property
    def bottom_face_x(self):
        """Bottom face in the X direction of the object.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.center[0]), face) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[0][1]
        except:
            return None

    @property
    def top_face_y(self):
        """Top face in the Y direction of the object.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.center[1]), face) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[-1][1]
        except:
            return None

    @property
    def bottom_face_y(self):
        """Bottom face in the X direction of the object.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.FacePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.center[1]), face) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[0][1]
        except:
            return None

    @property
    def top_edge_z(self):
        """Top edge in the Z direction of the object. Midpoint is used as criteria to find the edge.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        References
        ----------

        >>> oEditor.FaceCenter

        """
        try:
            result = [(float(face.top_edge_z.midpoint[2]), face.top_edge_z) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[-1][1]
        except:
            return None

    @property
    def bottom_edge_z(self):
        """Bottom edge in the Z direction of the object. Midpoint is used as criteria to find the edge.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        """
        try:
            result = [(float(face.bottom_edge_z.midpoint[2]), face.bottom_edge_z) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[0][1]
        except:
            return None

    @property
    def top_edge_x(self):
        """Top edge in the X direction of the object. Midpoint is used as criteria to find the edge.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        """
        try:
            result = [(float(face.top_edge_x.midpoint[0]), face.top_edge_x) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[-1][1]
        except:
            return None

    @property
    def bottom_edge_x(self):
        """Bottom edge in the X direction of the object. Midpoint is used as criteria to find the edge.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        """
        try:
            result = [(float(face.bottom_edge_x.midpoint[0]), face.bottom_edge_x) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[0][1]
        except:
            return None

    @property
    def top_edge_y(self):
        """Top edge in the Y direction of the object. Midpoint is used as criteria to find the edge.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        """
        try:
            result = [(float(face.top_edge_y.midpoint[1]), face.top_edge_y) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[-1][1]
        except:
            return None

    @property
    def bottom_edge_y(self):
        """Bottom edge in the X direction of the object. Midpoint is used as criteria to find the edge.

        Returns
        -------
        :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        """
        try:
            result = [(float(face.bottom_edge_y.midpoint[1]), face.bottom_edge_y) for face in self.faces]
            result = sorted(result, key=lambda tup: tup[0])
            return result[0][1]
        except:
            return None

    @property
    def edges(self):
        """Information for each edge in the given part.

        Returns
        -------
        list of :class:`pyaedt.modeler.elements3d.EdgePrimitive`

        References
        ----------

        >>> oEditor.GetEdgeIDsFromObject

        """
        if self.object_type == "Unclassified":
            return []
        edges = []
        for edge in self._primitives.get_object_edges(self.name):
            edge = int(edge)
            edges.append(EdgePrimitive(self, edge))
        return edges

    @property
    def vertices(self):
        """Information for each vertex in the given part.

        Returns
        -------
        list of :class:`pyaedt.modeler.elements3d.VertexPrimitive`

        References
        ----------

        >>> oEditor.GetVertexIDsFromObject

        """
        if self.object_type == "Unclassified":
            return []
        vertices = []

        v = [i for i in self._primitives.get_object_vertices(self.name)]
        if not v:
            for el in self.edges:
                pos = [float(p) for p in self._primitives.oeditor.GetEdgePositionAtNormalizedParameter(el.id, 0)]
                vertices.append(VertexPrimitive(self, -1, pos))
        if settings.aedt_version > "2022.2":
            v = v[::-1]
        for vertex in v:
            vertex = int(vertex)
            vertices.append(VertexPrimitive(self, vertex))
        return vertices

    @property
    def m_Editor(self):
        """Pointer to the oEditor object in the AEDT API. This property is
        intended primarily for use by FacePrimitive, EdgePrimitive, and
        VertexPrimitive child objects.

        Returns
        -------
        oEditor COM Object

        """
        return self._primitives.oeditor

    @property
    def logger(self):
        """Logger."""
        return self._primitives.logger

    @property
    def surface_material_name(self):
        """Surface material name of the object.

        Returns
        -------
        str or None
            Name of the surface material when successful, ``None`` and a warning message otherwise.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._surface_material is not None:
            return self._surface_material
        if "Surface Material" in self.valid_properties and self.model:
            self._surface_material = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Surface Material"
            )
            return self._surface_material.strip('"')

    @property
    def group_name(self):
        """Group the object belongs to.

        Returns
        -------
        str
            Name of the group.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._m_groupName is not None:
            return self._m_groupName
        if "Group" in self.valid_properties:
            self._m_groupName = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Group"
            )
            return self._m_groupName

    @group_name.setter
    def group_name(self, name):
        """Assign Object to a specific group. it creates a new group if the group doesn't exist.

        Parameters
        ----------
        name : str
            Name of the group to assign. Group will be created if it does not exist.

        Returns
        -------
        str
            Name of the group.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """

        if not list(self.m_Editor.GetObjectsInGroup(name)):
            self.m_Editor.CreateGroup(
                [
                    "NAME:GroupParameter",
                    "ParentGroupID:=",
                    "Model",
                    "Parts:=",
                    self._m_name,
                    "SubmodelInstances:=",
                    "",
                    "Groups:=",
                    "",
                ]
            )
            groupName = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Group"
            )
            self.m_Editor.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:Attributes",
                        ["NAME:PropServers", groupName],
                        ["NAME:ChangedProps", ["NAME:Name", "Value:=", name]],
                    ],
                ]
            )
            self._m_groupName = name
        else:
            vgroup = ["NAME:Group", "Value:=", name]
            self._change_property(vgroup)
            self._m_groupName = name

    @property
    def material_name(self):
        """Material name of the object.

        Returns
        -------
        str or None
            Name of the material when successful, ``None`` and a warning message otherwise.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._material_name is not None:
            return self._material_name
        if "Material" in self.valid_properties and self.model:
            mat = _retry_ntimes(10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Material")
            self._material_name = ""
            if mat:
                self._material_name = mat.strip('"').lower()
            return self._material_name
        return ""

    @material_name.setter
    def material_name(self, mat):
        matobj = self._primitives._materials.checkifmaterialexists(mat)
        if matobj:
            if not self.model:
                self.model = True
            vMaterial = ["NAME:Material", "Value:=", chr(34) + matobj.name + chr(34)]
            self._change_property(vMaterial)
            self._material_name = matobj.name.lower()
            self._solve_inside = None
        else:
            self.logger.warning("Material %s does not exist.", mat)

    @surface_material_name.setter
    def surface_material_name(self, mat):
        try:
            if not self.model:
                self.model = True
            self._surface_material = mat
            vMaterial = ["NAME:Surface Material", "Value:=", '"' + mat + '"']
            self._change_property(vMaterial)
            self._surface_material = mat
        except:
            self.logger.warning("Material %s does not exist", mat)

    @property
    def id(self):
        """ID of the object.

        Returns
        -------
        int or None
            ID of the object when successful, ``None`` otherwise.

        References
        ----------

        >>> oEditor.GetObjectIDByName

        """
        if not self._id:
            try:
                self._id = self._primitives.oeditor.GetObjectIDByName(self._m_name)
            except:
                return None
        return self._id

    @property
    def object_type(self):
        """Type of the object.

        Options are:
             * Solid
             * Sheet
             * Line

        Returns
        -------
        str
            Type of the object.

        """
        if self._object_type:
            return self._object_type
        if self._m_name in list(self.m_Editor.GetObjectsInGroup("Solids")):
            self._object_type = "Solid"
        elif self._m_name in list(self.m_Editor.GetObjectsInGroup("Sheets")):
            self._object_type = "Sheet"
        elif self._m_name in list(self.m_Editor.GetObjectsInGroup("Lines")):
            self._object_type = "Line"
        elif self._m_name in list(self.m_Editor.GetObjectsInGroup("Unclassified")):  # pragma: no cover
            self._object_type = "Unclassified"  # pragma: no cover
        return self._object_type

    @property
    def is3d(self):
        """Check for if the object is 3D.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if self.object_type == "Solid":
            return True
        else:
            return False

    @property
    def mass(self):
        """Object mass.

        Returns
        -------
        float or None
            Mass of the object when successful, 0.0 otherwise. Mass of the volume in kg since AEDT mass density is
            always in kg/m^3

        References
        ----------

        >>> oEditor.GetObjectVolume

        """
        if self.model and self.material_name:
            volume = self._primitives.oeditor.GetObjectVolume(self._m_name)
            units = self.object_units
            mass_density = (
                float(self._primitives._materials[self.material_name].mass_density.value)
                * float(volume)
                * float(AEDT_UNITS["Length"][str(units)]) ** 3
            )
            self._mass = mass_density
        else:
            self._mass = 0.0
        return self._mass

    @property
    def volume(self):
        """Object volume.

        Returns
        -------
        float
            Volume of the object when successful, 0.0 otherwise.

        References
        ----------

        >>> oEditor.GetObjectVolume

        """
        if self.object_type == "Solid":
            self._volume = float(self._primitives.oeditor.GetObjectVolume(self._m_name))
        else:
            self._volume = 0.0
        return self._volume

    @property
    def name(self):
        """Name of the object as a string value.

        Returns
        -------
        str
           Name of object as a string value.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        return self._m_name

    @name.setter
    def name(self, obj_name):
        if obj_name not in self._primitives.object_names:
            if obj_name != self._m_name:
                vName = []
                vName.append("NAME:Name")
                vName.append("Value:=")
                vName.append(obj_name)
                vChangedProps = ["NAME:ChangedProps", vName]
                vPropServers = ["NAME:PropServers"]
                vPropServers.append(self._m_name)
                vGeo3d = ["NAME:Geometry3DAttributeTab", vPropServers, vChangedProps]
                vOut = ["NAME:AllTabs", vGeo3d]
                _retry_ntimes(10, self._primitives.oeditor.ChangeProperty, vOut)
                self._m_name = obj_name
                self._primitives.cleanup_objects()
        else:
            pass

    @property
    def valid_properties(self):
        """Valid properties.

        References
        ----------

        >>> oEditor.GetProperties
        """
        if not self._all_props:
            self._all_props = _retry_ntimes(10, self.m_Editor.GetProperties, "Geometry3DAttributeTab", self._m_name)
        return self._all_props

    @property
    def color(self):
        """Part color as a tuple of integer values for `(Red, Green, Blue)` color values.

        If the integer values are outside the range 0-255, then limit the values. Invalid inputs are ignored.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        Examples
        --------
        >>> part.color = (255,255,0)

        """
        if self._color is not None:
            return self._color
        if "Color" in self.valid_properties:
            color = _retry_ntimes(10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Color")
            if color:
                b = (int(color) >> 16) & 255
                g = (int(color) >> 8) & 255
                r = int(color) & 255
                self._color = (r, g, b)
            else:
                self._color = (0, 195, 255)
            return self._color

    @property
    def color_string(self):
        """Color tuple as a string in the format '(Red, Green, Blue)'.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty
        """
        return "({} {} {})".format(self.color[0], self.color[1], self.color[2])

    @color.setter
    def color(self, color_value):
        color_tuple = None
        if isinstance(color_value, str):
            try:
                color_tuple = rgb_color_codes[color_value]
            except KeyError:
                parse_string = color_value.replace(")", "").replace("(", "").split()
                if len(parse_string) == 3:
                    color_tuple = tuple([int(x) for x in parse_string])
        else:
            try:
                color_tuple = tuple([int(x) for x in color_value])
            except ValueError:
                pass

        if color_tuple:
            try:
                R = clamp(color_tuple[0], 0, 255)
                G = clamp(color_tuple[1], 0, 255)
                B = clamp(color_tuple[2], 0, 255)
                vColor = ["NAME:Color", "R:=", str(R), "G:=", str(G), "B:=", str(B)]
                self._change_property(vColor)
                self._color = (R, G, B)
            except TypeError:
                color_tuple = None
        else:
            msg_text = "Invalid color input {} for object {}.".format(color_value, self._m_name)
            self._primitives.logger.warning(msg_text)

    @property
    def transparency(self):
        """Part transparency as a value between 0.0 and 1.0.

        If the value is outside the range, then apply a limit. If the value is not a valid number, set to ``0.0``.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._transparency is not None:
            return self._transparency
        if "Transparent" in self.valid_properties:
            transp = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Transparent"
            )
            try:
                self._transparency = float(transp)
            except:
                self._transparency = 0.3
            return self._transparency

    @transparency.setter
    def transparency(self, T):
        try:
            trans_float = float(T)
            if trans_float < 0.0:
                trans_float = 0.0
            elif trans_float > 1.0:
                trans_float = 1.0
        except ValueError:
            trans_float = 0.0
        vTrans = ["NAME:Transparent", "Value:=", str(trans_float)]

        self._change_property(vTrans)

        self._transparency = trans_float

    @property
    def object_units(self):
        """Object units."""
        return self._primitives.model_units

    @property
    def part_coordinate_system(self):
        """Part coordinate system.

        Returns
        -------
        str
            Name of the part coordinate system.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._part_coordinate_system is not None and not isinstance(self._part_coordinate_system, int):
            return self._part_coordinate_system
        if "Orientation" in self.valid_properties:
            self._part_coordinate_system = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Orientation"
            )
            return self._part_coordinate_system

    @part_coordinate_system.setter
    def part_coordinate_system(self, sCS):
        pcs = ["NAME:Orientation", "Value:=", sCS]
        self._change_property(pcs)
        self._part_coordinate_system = sCS
        return True

    @property
    def solve_inside(self):
        """Part solve inside flag.

        Returns
        -------
        bool
            ``True`` when ``"solve-inside"`` is activated for the part, ``False`` otherwise.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._solve_inside is not None:
            return self._solve_inside
        if "Solve Inside" in self.valid_properties and self.model:
            solveinside = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Solve Inside"
            )
            if solveinside == "false" or solveinside == "False":
                self._solve_inside = False
            else:
                self._solve_inside = True
            return self._solve_inside
        return None

    @solve_inside.setter
    def solve_inside(self, S):
        if not self.model:
            self.model = True
        vSolveInside = []
        # fS = self._to_boolean(S)
        fs = S
        vSolveInside.append("NAME:Solve Inside")
        vSolveInside.append("Value:=")
        vSolveInside.append(fs)
        self._change_property(vSolveInside)
        self._solve_inside = fs

    @property
    def display_wireframe(self):
        """Wireframe property of the part.

        Returns
        -------
        bool
            ``True`` when wirefame is activated for the part, ``False`` otherwise.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._wireframe is not None:
            return self._wireframe
        if "Display Wireframe" in self.valid_properties:
            wireframe = _retry_ntimes(
                10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Display Wireframe"
            )
            if wireframe == "true" or wireframe == "True":
                self._wireframe = True
            else:
                self._wireframe = False
            return self._wireframe

    @display_wireframe.setter
    def display_wireframe(self, fWireframe):
        vWireframe = ["NAME:Display Wireframe", "Value:=", fWireframe]
        # fwf = self._to_boolean(wf)

        self._change_property(vWireframe)
        self._wireframe = fWireframe

    @property
    def history(self):
        """Object history.

        Returns
        -------
            :class:`pyaedt.modeler.cad.elements3d.BinaryTree` when successful,
            ``False`` when failed.

        """
        try:
            child_object = self.m_Editor.GetChildObject(self.name)
            parent = BinaryTreeNode(self.name, child_object, True)
            return parent
        except:
            return False

    @property
    def model(self):
        """Part model or non-model property.

        Returns
        -------
        bool
            ``True`` when model, ``False`` otherwise.

        References
        ----------

        >>> oEditor.GetPropertyValue
        >>> oEditor.ChangeProperty

        """
        if self._model is not None:
            return self._model
        if "Model" in self.valid_properties:
            mod = _retry_ntimes(10, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, "Model")
            if mod == "false" or mod == "False":
                self._model = False
            else:
                self._model = True
            return self._model

    @model.setter
    def model(self, fModel):
        vArg1 = ["NAME:Model", "Value:=", fModel]
        fModel = _to_boolean(fModel)
        self._change_property(vArg1)
        self._model = fModel

    @pyaedt_function_handler()
    def unite(self, object_list):
        """Unite a list of objects with this object.

        Parameters
        ----------
        object_list : list of str or list of pyaedt.modeler.object3d.Object3d
            List of objects.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
           Object 3D object.

        References
        ----------

        >>> oEditor.Unite

        """
        unite_list = [self.name] + self._primitives.modeler.convert_to_selections(object_list, return_list=True)
        self._primitives.modeler.unite(unite_list)
        return self

    @pyaedt_function_handler()
    def intersect(self, theList, keep_originals=False):
        """Intersect the active object with a given list.

        Parameters
        ----------
        theList : list
            List of objects.
        keep_originals : bool, optional
            Whether to keep the original object. The default is ``False``.

        Returns
        -------
        :class:`pyaedt.modeler.object3d.Object3d`
            Retrieve the resulting 3D Object when succeeded.

        References
        ----------

        >>> oEditor.Intersect
        """
        theList = [self.name] + self._primitives.modeler.convert_to_selections(theList, return_list=True)
        self._primitives.modeler.intersect(theList, keep_originals)
        return self

    @pyaedt_function_handler()
    def split(self, plane, sides="Both"):
        """Split the active object.

        Parameters
        ----------
        plane : str
            Coordinate plane of the cut or the Application.PLANE object.
            Choices for the coordinate plane are ``"XY"``, ``"YZ"``, and ``"ZX"``.
        sides : str, optional
            Which side to keep. Options are ``"Both"``, ``"PositiveOnly"``,
            and ``"NegativeOnly"``. The default is ``"Both"``, in which case
            all objects are kept after the split.

        Returns
        -------
        list of str
            List of split objects.

        References
        ----------

        >>> oEditor.Split
        """
        return self._primitives.modeler.split(self.name, plane, sides)

    @pyaedt_function_handler()
    def mirror(self, position, vector, duplicate=False):
        """Mirror a selection.

        Parameters
        ----------
        position : list of int or float
            Cartesian ``[x, y, z]`` coordinates or
            the ``Application.Position`` object of a point in the plane used for the mirror operation.
        vector : list of float
            Vector in Cartesian coordinates ``[x1, y1, z1]``  or
            the ``Application.Position`` object for the vector normal to the plane used for the mirror operation.

        Returns
        -------
        pyaedt.modeler.Object3d.Object3d, bool
            3D object.
            ``False`` when failed.

        References
        ----------

        >>> oEditor.Mirror
        """
        if self._primitives.modeler.mirror(self.id, position=position, vector=vector, duplicate=duplicate):
            return self
        return False

    @pyaedt_function_handler()
    def rotate(self, cs_axis, angle=90.0, unit="deg"):
        """Rotate the selection.

        Parameters
        ----------
        cs_axis
            Coordinate system axis or the Application.CoordinateSystemAxis object.
        angle : float, optional
            Angle of rotation. The units, defined by ``unit``, can be either
            degrees or radians. The default is ``90.0``.
        unit : text, optional
             Units for the angle. Options are ``"deg"`` or ``"rad"``.
             The default is ``"deg"``.

        Returns
        -------
        pyaedt.modeler.Object3d.Object3d, bool
            3D object.
            ``False`` when failed.

        References
        ----------

        >>> oEditor.Rotate
        """
        if self._primitives.modeler.rotate(self.id, cs_axis=cs_axis, angle=angle, unit=unit):
            return self
        return False

    @pyaedt_function_handler()
    def move(self, vector):
        """Move objects from a list.

        Parameters
        ----------
        objid : list, Position object
            List of object IDs.
        vector : list
            Vector of the direction move. It can be a list of the ``[x, y, z]``
            coordinates or a Position object.

        Returns
        -------
        pyaedt.modeler.Object3d.Object3d, bool
            3D object.
            ``False`` when failed.

        References
        ----------
        >>> oEditor.Move
        """
        if self._primitives.modeler.move(self.id, vector=vector):
            return self
        return False

    def duplicate_around_axis(self, cs_axis, angle=90, nclones=2, create_new_objects=True):
        """Duplicate the object around the axis.

        Parameters
        ----------
        cs_axis : Application.CoordinateSystemAxis object
            Coordinate system axis of the object.
        angle : float
            Angle of rotation in degrees. The default is ``90``.
        nclones : int, optional
            Number of clones. The default is ``2``.
        create_new_objects : bool, optional
            Whether to create copies as new objects. The default is ``True``.

        Returns
        -------
        list
            List of names of the newly added objects.

        References
        ----------

        >>> oEditor.DuplicateAroundAxis

        """
        ret, added_objects = self._primitives.modeler.duplicate_around_axis(
            self, cs_axis, angle, nclones, create_new_objects
        )
        return added_objects

    @pyaedt_function_handler()
    def duplicate_along_line(self, vector, nclones=2, attachObject=False):
        """Duplicate the object along a line.

        Parameters
        ----------
        vector : list
            List of ``[x1 ,y1, z1]`` coordinates for the vector or the Application.Position object.
        nclones : int, optional
            Number of clones. The default is ``2``.
        attachObject : bool, optional
            Whether to attach the object. The default is ``False``.

        Returns
        -------
        list
            List of names of the newly added objects.

        References
        ----------

        >>> oEditor.DuplicateAlongLine

        """
        ret, added_objects = self._primitives.modeler.duplicate_along_line(self, vector, nclones, attachObject)
        return added_objects

    @pyaedt_function_handler()
    def translate(self, vector):
        """Translate the object and return the 3D object.

        .. deprecated:: 0.4.0
           Use :func:`move` instead.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
            3D object.

        References
        ----------

        >>> oEditor.Move

        """
        warnings.warn("`translate` is deprecated. Use `move` instead.", DeprecationWarning)
        self.move(vector)
        return self

    @pyaedt_function_handler()
    def sweep_along_vector(self, sweep_vector, draft_angle=0, draft_type="Round"):
        """Sweep the selection along a vector.

        Parameters
        ----------
        sweep_vector : list
            Application.Position object.
        draft_angle : float, optional
            Angle of the draft in degrees. The default is ``0``.
        draft_type : str, optional
            Type of the draft. Options are ``"Extended"``, ``"Round"``,
            and ``"Natural"``. The default value is ``"Round``.

        Returns
        -------
        bool
            ``True`` when model, ``False`` otherwise.

        References
        ----------

        >>> oEditor.SweepAlongVector

        """
        self._primitives.modeler.sweep_along_vector(self, sweep_vector, draft_angle, draft_type)
        return self

    @pyaedt_function_handler()
    def sweep_along_path(
        self, sweep_object, draft_angle=0, draft_type="Round", is_check_face_intersection=False, twist_angle=0
    ):
        """Sweep the selection along a vector.

        Parameters
        ----------
        sweep_object : :class:`pyaedt.modeler.object3d.Object3d`
            Application.Position object.
        draft_angle : float, optional
            Angle of the draft in degrees. The default is ``0``.
        draft_type : str
            Type of the draft. Options are ``"Extended"``, ``"Round"``,
            and ``"Natural"``. The default is ``"Round``.
        is_check_face_intersection : bool, optional
           The default value is ``False``.
        twist_angle : float, optional
            Angle at which to twist or rotate in degrees. The default value is ``0``.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
            Swept object.

        References
        ----------

        >>> oEditor.SweepAlongPath

        """
        self._primitives.modeler.sweep_along_path(
            self, sweep_object, draft_angle, draft_type, is_check_face_intersection, twist_angle
        )
        return self

    @pyaedt_function_handler()
    def sweep_around_axis(self, cs_axis, sweep_angle=360, draft_angle=0):
        """Sweep around an axis.

        Parameters
        ----------
        cs_axis : :class:`pyaedt.generic.constants.CoordinateSystemAxis`
            Coordinate system of the axis.
        sweep_angle : float, optional
             Sweep angle in degrees. The default is ``360``.
        draft_angle : float, optional
            Angle of the draft. The default is ``0``.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
            Swept object.

        References
        ----------

        >>> oEditor.SweepAroundAxis

        """
        self._primitives.modeler.sweep_around_axis(self, cs_axis, sweep_angle, draft_angle)
        return self

    @pyaedt_function_handler()
    def section(self, plane, create_new=True, section_cross_object=False):
        """Section the object.

        Parameters
        ----------
        plane : pyaedt.generic.constants.PLANE
            Coordinate system of the plane object. Application.PLANE object
        create_new : bool, optional
            Whether to create an object. The default is ``True``.
        section_cross_object : bool, optional
            The default is ``False``.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
            3D object.

        References
        ----------

        >>> oEditor.Section

        """
        self._primitives.modeler.section(self, plane, create_new, section_cross_object)
        return self

    @pyaedt_function_handler()
    def clone(self):
        """Clone the object and return the new 3D object.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
            3D object that was added.

        References
        ----------

        >>> oEditor.Clone

        """
        new_obj_tuple = self._primitives.modeler.clone(self.id)
        success = new_obj_tuple[0]
        assert success, "Could not clone the object {}.".format(self.name)
        new_name = new_obj_tuple[1][0]
        return self._primitives[new_name]

    @pyaedt_function_handler()
    def subtract(self, tool_list, keep_originals=True):
        """Subtract one or more parts from the object.

        Parameters
        ----------
        tool_list : str, Object3d, or list of str and Object3d.
            List of parts to subtract from this part.
        keep_originals : bool, optional
            Whether to keep the tool parts after subtraction. The default
            is ``True``. If ``False``, the parts are deleted.

        Returns
        -------
        pyaedt.modeler.object3d.Object3d
            Modified 3D object following the subtraction.

        References
        ----------

        >>> oEditor.Subtract

        """
        self._primitives.modeler.subtract(self.name, tool_list, keep_originals)
        return self

    @pyaedt_function_handler()
    def wrap_sheet(self, object_name, imprinted=False):
        """Execute the sheet wrapping around an object. This object can be either the sheet or the object.
        If wrapping produces an unclassified operation it will be reverted.

        Parameters
        ----------
        object_name : str, :class:`pyaedt.modeler.Object3d.Object3d`
            Object name or solid object or sheet name.
        imprinted : bool, optional
            Either if imprint or not over the sheet. Default is `False`.

        Returns
        -------
        bool
            Command execution status.
        """
        object_name = self._primitives.convert_to_selections(object_name, False)
        if self.object_type == "Sheet" and object_name in self._primitives.solid_names:
            return self._primitives.wrap_sheet(self.name, object_name, imprinted)
        elif self.object_type == "Solid" and object_name in self._primitives.sheet_names:
            return self._primitives.wrap_sheet(object_name, self.name, imprinted)
        else:
            msg = "Error in command execution."
            msg += " Either one of the two objects has to be a sheet and the other an object."
            self.logger.error(msg)
            return False

    @pyaedt_function_handler()
    def delete(self):
        """Delete the object.

        References
        ----------

        >>> oEditor.Delete
        """
        arg = ["NAME:Selections", "Selections:=", self._m_name]
        self.m_Editor.Delete(arg)
        self._primitives.cleanup_objects()
        self.__dict__ = {}

    @pyaedt_function_handler()
    def faces_by_area(self, area, area_filter="==", tolerance=1e-12):
        """Filter faces by area.

        Parameters
        ----------
        area : float
            Value of the area to filter in model units.
        area_filter : str, optional
            Comparer symbol.
            Default value is "==".
        tolerance : float, optional
            tolerance for comparison.

        Returns
        -------
        list of :class:`pyaedt.modeler.elements3d.FacePrimitive`
            List of face primitives.
        """

        filters = ["==", "<=", ">=", "<", ">"]
        if area_filter not in filters:
            raise ValueError('Symbol not valid, enter one of the following: "==", "<=", ">=", "<", ">"')

        faces = []
        for face in self.faces:
            if area_filter == "==":
                if abs(face.area - area) < tolerance:
                    faces.append(face)
            if area_filter == ">=":
                if (face.area - area) >= -tolerance:
                    faces.append(face)
            if area_filter == "<=":
                if (face.area - area) <= tolerance:
                    faces.append(face)
            if area_filter == ">":
                if (face.area - area) > 0:
                    faces.append(face)
            if area_filter == "<":
                if (face.area - area) < 0:
                    faces.append(face)

        return faces

    @pyaedt_function_handler()
    def edges_by_length(self, length, length_filter="==", tolerance=1e-12):
        """Filter edges by length.

        Parameters
        ----------
        length : float
            Value of the length to filter.
        length_filter : str, optional
            Comparer symbol.
            Default value is "==".
        tolerance : float, optional
            tolerance for comparison.

        Returns
        -------
        list of :class:`pyaedt.modeler.elements3d.EdgePrimitive`
            List of edge primitives.
        """
        filters = ["==", "<=", ">=", "<", ">"]
        if length_filter not in filters:
            raise ValueError('Symbol not valid, enter one of the following: "==", "<=", ">=", "<", ">"')

        edges = []
        for edge in self.edges:
            if length_filter == "==":
                if abs(edge.length - length) < tolerance:
                    edges.append(edge)
            if length_filter == ">=":
                if (edge.length - length) >= -tolerance:
                    edges.append(edge)
            if length_filter == "<=":
                if (edge.length - length) <= tolerance:
                    edges.append(edge)
            if length_filter == ">":
                if (edge.length - length) > 0:
                    edges.append(edge)
            if length_filter == "<":
                if (edge.length - length) < 0:
                    edges.append(edge)

        return edges

    @pyaedt_function_handler()
    def _change_property(self, vPropChange):
        return self._primitives._change_geometry_property(vPropChange, self._m_name)

    def _update(self):
        # self._object3d._refresh_object_types()
        self._primitives.cleanup_objects()

    def __str__(self):
        return """
         {}
         name: {}    id: {}    object_type: {}
         --- read/write properties  ----
         solve_inside: {}
         model: {}
         material_name: {}
         color: {}
         transparency: {}
         display_wireframe {}
         part_coordinate_system: {}
         """.format(
            type(self),
            self.name,
            self.id,
            self.object_type,
            self.solve_inside,
            self.model,
            self.material_name,
            self.color,
            self.transparency,
            self.display_wireframe,
            self.part_coordinate_system,
        )



# Power Sinusoidal
PowerSinusoidal = [
    "NAME:Name",
    "DataId:=",
    "",
    "Type:=",
    1,
    "Output:=",
    2,
    "NumPins:=",
    2,
    "Netlist:=",
    "",
    "CompName:=",
    "Nexxim Circuit Elements\\Independent Sources:P_SIN",
    "FDSFileName:=",
    "",
    "BtnPropFileName:=",
    "",
    [
        "NAME:Properties",
        "TextProp:=",
        ["LabelID", "HD", "Property string for netlist ID", "V@ID"],
        "ValueProp:=",
        ["ACMAG", "D", "AC magnitude for small-signal analysis (Volts)", "nan V", 0],
        "ValuePropNU:=",
        ["ACPHASE", "D", "AC phase for small-signal analysis", "0deg", 0, "deg", "AdditionalPropInfo:=", ""],
        "ValueProp:=",
        ["DC", "D", "DC voltage (Volts)", "0V", 0],
        "ValuePropNU:=",
        ["VO", "D", "Power offset from zero watts", "0W", 0, "W", "AdditionalPropInfo:=", ""],
        "ValueProp:=",
        ["POWER", "D", "Available power of the source above VO", "0W", 0],
        "ValueProp:=",
        ["FREQ", "D", "Frequency (Hz)", "1GHz", 0],
        "ValueProp:=",
        ["TD", "D", "Delay to start of sine wave (seconds)", "0s", 0],
        "ValueProp:=",
        ["ALPHA", "D", "Damping factor (1/seconds)", "0", 0],
        "ValuePropNU:=",
        ["THETA", "D", "Phase delay", "0deg", 0, "deg", "AdditionalPropInfo:=", ""],
        "ValueProp:=",
        [
            "TONE",
            "D",
            "Frequency (Hz) to use for harmonic balance analysis, should be a "
            "submultiple of (or equal to) the driving frequency and should also be "
            "included in the HB analysis setup",
            "0Hz",
            0,
        ],
        "TextProp:=",
        ["ModelName", "SHD", "", "P_SIN"],
        "ButtonProp:=",
        ["CosimDefinition", "D", "", "Edit", "Edit", 40501, "ButtonPropClientData:=", []],
        "MenuProp:=",
        ["CoSimulator", "D", "", "DefaultNetlist", 0],
        [
            "Netlist",
            0,
            0,
            "V@ID %0 %1 *DC(DC=@DC) POWER SIN(?VO(@VO) ?POWER(@POWER) ?FREQ(@FREQ) ?TD(@TD) "
            "?ALPHA(@ALPHA) ?THETA(@THETA)) *TONE(TONE=@TONE) *ACMAG(AC @ACMAG @ACPHASE)",
        ],
    ],
]

PSinProps = [
    "ACMAG",
    "ACPHASE",
    "DC",
    "VO",
    "POWER",
    "FREQ",
    "TD",
    "ALPHA",
    "THETA",
    "TONE",
    "CosimDefinition",
    "CoSimulator",
    "CoSimulator/Choices",
    "Netlist",
]

# Voltage Sinusoidal
VoltageSinusoidal = [
    "NAME:Name",
    "DataId:=",
    "",
    "Type:=",
    1,
    "Output:=",
    2,
    "NumPins:=",
    2,
    "Netlist:=",
    "",
    "CompName:=",
    "Nexxim Circuit Elements\\Independent Sources:V_SIN",
    "FDSFileName:=",
    "",
    "BtnPropFileName:=",
    "",
    [
        "NAME:Properties",
        "TextProp:=",
        ["LabelID", "HD", "Property string for netlist ID", "V@ID"],
        "ValueProp:=",
        ["ACMAG", "D", "AC magnitude for small-signal analysis (Volts)", "nan V", 0],
        "ValuePropNU:=",
        ["ACPHASE", "D", "AC phase for small-signal analysis", "0deg", 0, "deg"],
        "ValueProp:=",
        ["DC", "D", "DC voltage (Volts)", "0V", 0],
        "ValueProp:=",
        ["VA", "D", "Voltage amplitude (Volts)", "0V", 0],
        "ValuePropNU:=",
        ["VO", "D", "Voltage offset from zero (Volts)", "0V", 0, "V"],
        "ValueProp:=",
        ["FREQ", "D", "Frequency (Hz)", "1GHz", 0],
        "ValueProp:=",
        ["TD", "D", "Delay to start of sine wave (seconds)", "0s", 0],
        "ValueProp:=",
        ["ALPHA", "D", "Damping factor (1/seconds)", "0", 0],
        "ValuePropNU:=",
        ["THETA", "D", "Phase delay", "0deg", 0, "deg"],
        "ValueProp:=",
        [
            "TONE",
            "D",
            "Frequency (Hz) to use for harmonic balance analysis, should be a "
            "submultiple of (or equal to) the driving frequency and should also be "
            "included in the HB analysis setup",
            "0Hz",
            0,
        ],
        "TextProp:=",
        ["ModelName", "SHD", "", "V_SIN"],
        "ButtonProp:=",
        ["CosimDefinition", "D", "", "", "Edit", 40501, "ButtonPropClientData:=", []],
        "MenuProp:=",
        ["CoSimulator", "D", "", "DefaultNetlist", 0],
        [
            "Netlist",
            0,
            0,
            "V@ID %0 %1 *DC(DC=@DC) SIN(?VO(@VO) ?VA(@VA) ?FREQ(@FREQ) ?TD(@TD) ?ALPHA(@ALPHA) ?THETA(@THETA))"
            " *TONE(TONE=@TONE) *ACMAG(AC @ACMAG @ACPHASE)",
        ],
    ],
]

VSinProps = [
    "ACMAG",
    "ACPHASE",
    "DC",
    "VA",
    "VO",
    "FREQ",
    "TD",
    "ALPHA",
    "THETA",
    "TONE",
    "CosimDefinition",
    "CoSimulator",
    "CoSimulator/Choices",
    "Netlist",
]

# Current Sinusoidal
CurrentSinusoidal = [
    "NAME:Name",
    "DataId:=",
    "",
    "Type:=",
    1,
    "Output:=",
    2,
    "NumPins:=",
    2,
    "Netlist:=",
    "",
    "CompName:=",
    "Nexxim Circuit Elements\\Independent Sources:I_SIN",
    "FDSFileName:=",
    "",
    "BtnPropFileName:=",
    "",
    [
        "NAME:Properties",
        "TextProp:=",
        ["LabelID", "HD", "Property string for netlist ID", "I@ID"],
        "ValueProp:=",
        ["ACMAG", "D", "AC magnitude for small-signal analysis (Amps)", "nan A", 0],
        "ValuePropNU:=",
        ["ACPHASE", "D", "AC phase for small-signal analysis", "0deg", 0, "deg", "AdditionalPropInfo:=", ""],
        "ValueProp:=",
        ["DC", "D", "DC current (Amps)", "0A", 0],
        "ValuePropNU:=",
        ["VO", "D", "Current offset (Amps)", "0A", 0, "A", "AdditionalPropInfo:=", ""],
        "ValueProp:=",
        ["VA", "D", "Current amplitude (Amps)", "0A", 0],
        "ValueProp:=",
        ["FREQ", "D", "Frequency (Hz)", "1GHz", 0],
        "ValueProp:=",
        ["TD", "D", "Delay to start of sine wave (seconds)", "0s", 0],
        "ValueProp:=",
        ["ALPHA", "D", "Damping factor (1/seconds)", "0", 0],
        "ValuePropNU:=",
        ["THETA", "D", "Phase delay", "0deg", 0, "deg"],
        "ValueProp:=",
        ["M", "D", "Multiplier for simulating multiple parallel current sources", "1", 0],
        "ValueProp:=",
        [
            "TONE",
            "D",
            "Frequency (Hz) to use for harmonic balance analysis, should be a "
            "submultiple of (or equal to) the driving frequency and should also be "
            "included in the HB analysis setup",
            "0Hz",
            0,
        ],
        "TextProp:=",
        ["ModelName", "SHD", "", "I_SIN"],
        "ButtonProp:=",
        ["CosimDefinition", "D", "", "", "Edit", 40501, "ButtonPropClientData:=", []],
        "MenuProp:=",
        ["CoSimulator", "D", "", "DefaultNetlist", 0],
        [
            "Netlist",
            0,
            0,
            "I@ID %0 %1 *DC(DC=@DC) SIN(?VO(@VO) ?VA(@VA) ?FREQ(@FREQ) ?TD(@TD) ?ALPHA(@ALPHA) ?THETA(@THETA) *M(M=@M))"
            "*TONE(TONE=@TONE) *ACMAG(AC @ACMAG @ACPHASE)",
        ],
    ],
]

ASinProps = [
    "ACMAG",
    "ACPHASE",
    "DC",
    "VO",
    "VA",
    "FREQ",
    "TD",
    "ALPHA",
    "THETA",
    "M",
    "TONE",
    "CosimDefinition",
    "CoSimulator",
    "CoSimulator/Choices",
    "Netlist",
]


# Power IQ
iq_properties = [
    "NAME:Properties",
    "TextProp:=",
    ["LabelID", "HD", "Property String for netlist ID", "A@ID"],
    "ValueProp:=",
    ["FC", "D", "Carrier frequency (Hertz)", "1GHz", 0],
    "ValueProp:=",
    ["TS", "D", "Sampling time (Second)", "nan s", 0],
    "ValueProp:=",
    ["DC", "D", "Source value at DC (Volt)", "0V", 0],
    "ValueProp:=",
    ["R", "D", "Repeat from time (Second)", "nan s", 0],
    "ValueProp:=",
    ["TD", "D", "Time delay (Second)", "0s", 0],
    "ValueProp:=",
    ["V", "D", "Carrier amplitude, voltage-based (Volts)", "1V", 0],
    "ValueProp:=",
    ["VA", "D", "Carrier amplitude, power-based (Watts)", "nan W", 0],
    "ValueProp:=",
    ["VO", "D", "Carrier amplitude offset (Volts or Watts)", "0", 0],
    "ValueProp:=",
    ["RZ", "D", "Real carrier impedance (Ohm)", "50ohm", 0],
    "ValueProp:=",
    ["IZ", "D", "Imaginary carrier impedance (Ohm)", "0ohm", 0],
    "ValueProp:=",
    ["ALPHA", "D", "Damping factor (1/Seconds)", "0", 0],
    "ValuePropNU:=",
    ["THETA", "D", "Phase delay", "0deg", 0, "deg"],
    "ValueProp:=",
    ["TONE", "D", "Source tone for HB analysis", "nan ", 0],
]

for i in range(1, 21):
    iq_properties.append("ValueProp:=")
    iq_properties.append(
        ["time" + str(i), "D", "Timepoint value where the corresponding voltage values are valid", "0s", 0]
    )
    iq_properties.append("ValueProp:=")
    iq_properties.append(["ival" + str(i), "D", "I value at the corresponding timepoint (Volts)", "0V", 0])
    iq_properties.append("ValueProp:=")
    iq_properties.append(["qval" + str(i), "D", "Q value at the corresponding timepoint (Volts)", "0V", 0])

power_iq_properties = iq_properties + [
    "TextProp:=",
    ["COMPONENT", "RHD", "", "vsiq_source"],
    "MenuProp:=",
    ["CoSimulator", "D", "", "DefaultNetlist", 0],
    "ButtonProp:=",
    ["CosimDefinition", "D", "", "", "Edit", 40501, "ButtonPropClientData:=", []],
    "ButtonProp:=",
    ["file", "OD", "", "", "", 4, "ButtonPropClientData:=", []],
    "ButtonProp:=",
    ["IQData", "OD", "", "IQData", "IQData", 2, "ButtonPropClientData:=", []],
    [
        "Netlist",
        0,
        0,
        "A@ID %0  %1 *FC(FC=@FC) *TS(TS=@TS) *DC(DC=@DC) *R(R=@R) *TD(TD=@TD) *V(V=@V) *VA("
        "VA=@VA) "
        "*VO(VO=@VO) *RZ(RZ=@RZ) *IZ(IZ=@IZ) *ALPHA(ALPHA=@ALPHA) *THETA(THETA=@THETA) "
        "*TONE(TONE=@TONE) *FILE(FILE=@FILE) TIMES=[?time1(@time1) *time2(@time2) *time3("
        "@time3) *time4(@time4) *time5(@time5) *time6(@time6) *time7(@time7) *time8(@time8) "
        "*time9( @time9) *time10(@time10) *time11(@time11) *time12(@time12) *time13(@time13) "
        "*time14(@time14) *time15(@time15) *time16(@time16) *time17(@time17) *time18(@time18)"
        "*time19(@time19) *time20(@time20)] I_T=[?ival1(@ival1) *ival2(@ival2) *ival3("
        "@ival3) *ival4(@ival4) *ival5(@ival5) *ival6(@ival6) *ival7(@ival7) *ival8(@ival8) "
        "*ival9(@ival9) *ival10(@ival10) *ival11(@ival11) *ival12(@ival12) *ival13(@ival13) "
        "*ival14(@ival14) *ival15(@ival15) *ival16(@ival16) *ival17(@ival17) *ival18("
        "@ival18) *ival19(@ival19) *ival20(@ival20)] Q_T=[?qval1(@qval1) *qval2(@qval2) "
        "*qval3(@qval3) *qval4(@qval4) *qval5(@qval5) *qval6(@qval6) *qval7(@qval7) *qval8("
        "@qval8) *qval9(@qval9) *qval10(@qval10) *qval11(@qval11) *qval12(@qval12) *qval13("
        "@qval13) *qval14(@qval14) *qval15(@qval15) *qval16(@qval16) *qval17(@qval17) "
        "*qval18(@qval18) *qval19(@qval19) *qval20(@qval20)] COMPONENT=@COMPONENT",
    ],
]

PowerIQ = [
    "NAME:Name",
    "DataId:=",
    "",
    "Type:=",
    5,
    "Output:=",
    2,
    "NumPins:=",
    2,
    "Netlist:=",
    "",
    "CompName:=",
    "Nexxim Circuit Elements\\Independent Sources:VSIQ",
    "FDSFileName:=",
    "",
    "BtnPropFileName:=",
    "",
    power_iq_properties,
]

PowerIQProps = [
    "FC",
    "TS",
    "DC",
    "R",
    "TD",
    "V",
    "VA",
    "VO",
    "RZ",
    "IZ",
    "ALPHA",
    "THETA",
    "TONE",
    "time1",
    "ival1",
    "qval1",
    "time2",
    "ival2",
    "qval2",
    "time3",
    "ival3",
    "qval3",
    "time4",
    "ival4",
    "qval4",
    "time5",
    "ival5",
    "qval5",
    "time6",
    "ival6",
    "qval6",
    "time7",
    "ival7",
    "qval7",
    "time8",
    "ival8",
    "qval8",
    "time9",
    "ival9",
    "qval9",
    "time10",
    "ival10",
    "qval10",
    "time11",
    "ival11",
    "qval11",
    "time12",
    "ival12",
    "qval12",
    "time13",
    "ival13",
    "qval13",
    "time14",
    "ival14",
    "qval14",
    "time15",
    "ival15",
    "qval15",
    "time16",
    "ival16",
    "qval16",
    "time17",
    "ival17",
    "qval17",
    "time18",
    "ival18",
    "qval18",
    "time19",
    "ival19",
    "qval19",
    "time20",
    "ival20",
    "qval20",
    "CoSimulator",
    "CoSimulator/Choices",
    "CosimDefinition",
    "file",
    "IQData",
    "Netlist",
]

# Voltage Frequency dependent
VoltageFrequency = [
    "NAME:Name",
    "DataId:=",
    "",
    "Type:=",
    17,
    "Output:=",
    0,
    "NumPins:=",
    2,
    "Netlist:=",
    "",
    "CompName:=",
    "Nexxim Circuit Elements\\Independent Sources:V_PWL_F",
    "FDSFileName:=",
    "",
    "BtnPropFileName:=",
    "",
    [
        "NAME:Properties",
        "TextProp:=",
        ["LabelID", "HD", "Property string for netlist ID", "V@ID"],
        "ButtonProp:=",
        ["FreqDependentSourceData", "D", "Frequency Dependent Source Data", "", "", 7, "ButtonPropClientData:=", []],
        "TextProp:=",
        ["COMPONENT", "RHD", "", "vsource_freq"],
        "ButtonProp:=",
        ["CosimDefinition", "SD", "", "Edit", "Edit", 0, "ButtonPropClientData:=", []],
        "MenuProp:=",
        ["CoSimulator", "D", "", "DefaultNetlist", 0],
        ["Netlist", 0, 0, "V@ID %0 %1 *FreqDependentSourceData(@FreqDependentSourceData)"],
    ],
]

VoltageFrequencyProps = [
    "frequencies",
    "vmag",
    "vang",
    "vreal",
    "vimag",
    "fds_filename",
    "magnitude_angle",
    "FreqDependentSourceData",
    "CosimDefinition",
    "CoSimulator",
    "CoSimulator/Choices",
    "Netlist",
]

# Voltage DC
VoltageDC = [
    "NAME:Name",
    "DataId:=",
    "",
    "Type:=",
    1,
    "Output:=",
    2,
    "NumPins:=",
    2,
    "Netlist:=",
    "",
    "CompName:=",
    "Nexxim Circuit Elements\\Independent Sources:V_DC",
    "FDSFileName:=",
    "",
    "BtnPropFileName:=",
    "",
    [
        "NAME:Properties",
        "TextProp:=",
        ["LabelID", "HD", "Property string for netlist ID", "V@ID"],
        "ValueProp:=",
        ["ACMAG", "D", "AC magnitude for small-signal analysis (Volts)", "nan V", 0],
        "ValuePropNU:=",
        ["ACPHASE", "D", "AC phase for small-signal analysis", "0deg", 0, "deg"],
        "ValueProp:=",
        ["DC", "D", "DC voltage (Volts)", "0V", 0],
        "TextProp:=",
        ["ModelName", "SHD", "", "V_DC"],
        "MenuProp:=",
        ["CoSimulator", "D", "", "DefaultNetlist", 0],
        "ButtonProp:=",
        ["CosimDefinition", "D", "", "", "Edit", 40501, "ButtonPropClientData:=", []],
        "ButtonProp:=",
        ["Noise", "OD", "", "Noise", "Noise", 0, "ButtonPropClientData:=", []],
        "TextProp:=",
        ["ModelName", "SHD", "", "V_DC"],
        "ButtonProp:=",
        ["CosimDefinition", "D", "", "Edit", "Edit", 40501, "ButtonPropClientData:=", []],
        "MenuProp:=",
        ["CoSimulator", "D", "", "DefaultNetlist", 0],
        [
            "Netlist",
            0,
            0,
            "V@ID %0 %1 *DC(DC=@DC) *ACMAG(AC @ACMAG @ACPHASE)",
        ],
    ],
]

DCProps = [
    "ACMAG",
    "ACPHASE",
    "DC",
    "POWER",
    "CosimDefinition",
    "CoSimulator",
    "CoSimulator/Choices",
    "Noise",
    "Netlist",
]


class SourceKeys(object):
    """Provides source keys."""

    SourceTemplates = {
        "PowerSin": PowerSinusoidal,
        "PowerIQ": PowerIQ,
        "VoltageFrequencyDependent": VoltageFrequency,
        "VoltageDC": VoltageDC,
        "VoltageSin": VoltageSinusoidal,
        "CurrentSin": CurrentSinusoidal,
    }

    SourceProps = {
        "PowerSin": PSinProps,
        "PowerIQ": PowerIQProps,
        "VoltageFrequencyDependent": VoltageFrequencyProps,
        "VoltageDC": DCProps,
        "VoltageSin": VSinProps,
        "CurrentSin": ASinProps,
    }

    SourceNames = [
        "PowerSin",
        "PowerIQ",
        "VoltageFrequencyDependent",
        "VoltageDC",
        "VoltageSin",
        "CurrentSin",
    ]



import os.path

from pyaedt.edb_core.ipc2581.bom.bom import Bom
from pyaedt.edb_core.ipc2581.bom.bom_item import BomItem
from pyaedt.edb_core.ipc2581.content.content import Content
from pyaedt.edb_core.ipc2581.ecad.cad_data.padstack_def import PadstackDef
from pyaedt.edb_core.ipc2581.ecad.ecad import Ecad
from pyaedt.edb_core.ipc2581.history_record import HistoryRecord
from pyaedt.edb_core.ipc2581.logistic_header import LogisticHeader
from pyaedt.generic.general_methods import ET
from pyaedt.generic.general_methods import pyaedt_function_handler


class Ipc2581(object):
    """Manages the Ipc2581 exporter."""

    def __init__(self, pedb, units):
        self.revision = "C"
        self._pedb = pedb
        self.units = units
        self.content = Content(self)
        self.logistic_header = LogisticHeader()
        self.history_record = HistoryRecord()
        self.bom = Bom()
        self.ecad = Ecad(self, pedb, units)
        self.file_path = ""
        self.design_name = ""
        self.top_bottom_layers = []

    @pyaedt_function_handler()
    def load_ipc_model(self):
        self.design_name = self._pedb.cell_names[0]
        self.content.step_ref = self.design_name
        self._pedb.logger.info("Parsing Layers...")
        self.add_layers_info()
        self._pedb.logger.info("Parsing BOM...")
        self.add_bom()
        self._pedb.logger.info("Parsing Padstack Definitions...")
        self.add_pdstack_definition()
        self._pedb.logger.info("Parsing Components...")
        self.add_components()
        self._pedb.logger.info("Parsing Logical Nets...")
        self.add_logical_nets()
        self._pedb.logger.info("Parsing Layout Primitives...")
        self.add_layer_features()
        self._pedb.logger.info("Parsing Drills...")
        self.add_drills()
        self._pedb.logger.info("Parsing EDB Completed!")

    @pyaedt_function_handler()
    def add_pdstack_definition(self):
        for padstack_name, padstackdef in self._pedb.core_padstack.padstacks.items():
            padstack_def = PadstackDef()
            padstack_def.name = padstack_name
            padstack_def.padstack_hole_def.name = padstack_name
            if padstackdef.hole_properties:
                padstack_def.padstack_hole_def.diameter = self.from_meter_to_units(
                    padstackdef.hole_properties[0], self.units
                )
            for layer, pad in padstackdef.pad_by_layer.items():
                if pad.parameters_values:
                    if pad.geometry_type == 1:
                        primitive_ref = "CIRCLE_{}".format(
                            self.from_meter_to_units(pad.parameters_values[0], self.units)
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_circ_dict:
                            self.content.standard_geometries_dict.standard_circ_dict[
                                primitive_ref
                            ] = self.from_meter_to_units(pad.parameters_values[0], self.units)
                    elif pad.geometry_type == 2:
                        primitive_ref = "RECT_{}_{}".format(
                            self.from_meter_to_units(pad.parameters_values[0], self.units),
                            self.from_meter_to_units(pad.parameters_values[0], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_rect_dict:
                            self.content.standard_geometries_dict.standard_rect_dict[primitive_ref] = [
                                self.from_meter_to_units(pad.parameters_values[0], self.units),
                                self.from_meter_to_units(pad.parameters_values[0], self.units),
                            ]
                    elif pad.geometry_type == 3:
                        primitive_ref = "RECT_{}_{}".format(
                            self.from_meter_to_units(pad.parameters_values[0], self.units),
                            self.from_meter_to_units(pad.parameters_values[1], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_rect_dict:
                            self.content.standard_geometries_dict.standard_rect_dict[primitive_ref] = [
                                self.from_meter_to_units(pad.parameters_values[0], self.units),
                                self.from_meter_to_units(pad.parameters_values[1], self.units),
                            ]

                    elif pad.geometry_type == 4:
                        primitive_ref = "OVAL_{}_{}_{}".format(
                            self.from_meter_to_units(pad.parameters_values[0], self.units),
                            self.from_meter_to_units(pad.parameters_values[1], self.units),
                            self.from_meter_to_units(pad.parameters_values[2], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_oval_dict:
                            self.content.standard_geometries_dict.standard_oval_dict[primitive_ref] = [
                                self.from_meter_to_units(pad.parameters_values[0], self.units),
                                self.from_meter_to_units(pad.parameters_values[1], self.units),
                                self.from_meter_to_units(pad.parameters_values[2], self.units),
                            ]
                    else:
                        primitive_ref = "Default"
                    padstack_def.add_padstack_pad_def(layer=layer, pad_use="REGULAR", primitive_ref=primitive_ref)
            for layer, antipad in padstackdef.antipad_by_layer.items():
                if antipad.parameters_values:
                    if antipad.geometry_type == 1:
                        primitive_ref = "CIRCLE_{}".format(
                            self.from_meter_to_units(antipad.parameters_values[0], self.units)
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_circ_dict:
                            self.content.standard_geometries_dict.standard_circ_dict[
                                primitive_ref
                            ] = self.from_meter_to_units(antipad.parameters_values[0], self.units)
                    elif antipad.geometry_type == 2:
                        primitive_ref = "RECT_{}_{}".format(
                            self.from_meter_to_units(antipad.parameters_values[0], self.units),
                            self.from_meter_to_units(antipad.parameters_values[0], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_rect_dict:
                            self.content.standard_geometries_dict.standard_rect_dict[primitive_ref] = [
                                self.from_meter_to_units(antipad.parameters_values[0], self.units),
                                self.from_meter_to_units(antipad.parameters_values[0], self.units),
                            ]
                    elif antipad.geometry_type == 3:
                        primitive_ref = "RECT_{}_{}".format(
                            self.from_meter_to_units(antipad.parameters_values[0], self.units),
                            self.from_meter_to_units(antipad.parameters_values[1], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_rect_dict:
                            self.content.standard_geometries_dict.standard_rect_dict[primitive_ref] = [
                                self.from_meter_to_units(antipad.parameters_values[0], self.units),
                                self.from_meter_to_units(antipad.parameters_values[1], self.units),
                            ]
                    elif antipad.geometry_type == 4:
                        primitive_ref = "OVAL_{}_{}_{}".format(
                            self.from_meter_to_units(antipad.parameters_values[0], self.units),
                            self.from_meter_to_units(antipad.parameters_values[1], self.units),
                            self.from_meter_to_units(antipad.parameters_values[2], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_oval_dict:
                            self.content.standard_geometries_dict.standard_oval_dict[primitive_ref] = [
                                self.from_meter_to_units(antipad.parameters_values[0], self.units),
                                self.from_meter_to_units(antipad.parameters_values[1], self.units),
                                self.from_meter_to_units(antipad.parameters_values[2], self.units),
                            ]
                    else:
                        primitive_ref = "Default"
                    padstack_def.add_padstack_pad_def(layer=layer, pad_use="ANTIPAD", primitive_ref=primitive_ref)
            for layer, thermalpad in padstackdef.thermalpad_by_layer.items():
                if thermalpad.parameters_values:
                    if thermalpad.geometry_type == 1:
                        primitive_ref = "CIRCLE_{}".format(
                            self.from_meter_to_units(thermalpad.parameters_values[0], self.units)
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_circ_dict:
                            self.content.standard_geometries_dict[primitive_ref] = self.from_meter_to_units(
                                thermalpad.parameters_values[0], self.units
                            )
                    elif thermalpad.geometry_type == 2:
                        primitive_ref = "RECT_{}_{}".format(
                            self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                            self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_rect_dict:
                            self.content.standard_geometries_dict.standard_rect_dict[primitive_ref] = [
                                self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                                self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                            ]
                    elif thermalpad.geometry_type == 3:
                        primitive_ref = "RECT_{}_{}".format(
                            self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                            self.from_meter_to_units(thermalpad.parameters_values[1], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_rect_dict:
                            self.content.standard_geometries_dict.standard_rect_dict[primitive_ref] = [
                                self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                                self.from_meter_to_units(thermalpad.parameters_values[1], self.units),
                            ]
                    elif thermalpad.geometry_type == 4:
                        primitive_ref = "OVAL_{}_{}_{}".format(
                            self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                            self.from_meter_to_units(thermalpad.parameters_values[1], self.units),
                            self.from_meter_to_units(thermalpad.parameters_values[2], self.units),
                        )
                        if not primitive_ref in self.content.standard_geometries_dict.standard_oval_dict:
                            self.content.standard_geometries_dict.standard_oval_dict[primitive_ref] = [
                                self.from_meter_to_units(thermalpad.parameters_values[0], self.units),
                                self.from_meter_to_units(thermalpad.parameters_values[1], self.units),
                                self.from_meter_to_units(thermalpad.parameters_values[2], self.units),
                            ]
                    else:
                        primitive_ref = "Default"
                    padstack_def.add_padstack_pad_def(layer=layer, pad_use="THERMAL", primitive_ref=primitive_ref)
            if not padstack_def.name in self.ecad.cad_data.cad_data_step.padstack_defs:
                self.ecad.cad_data.cad_data_step.padstack_defs[padstack_def.name] = padstack_def

    @pyaedt_function_handler()
    def add_bom(self):
        # Bom
        for part_name, components in self._pedb.core_components.components_by_partname.items():
            bom_item = BomItem()
            bom_item.part_name = part_name
            bom_item.quantity = len(components)
            bom_item.pin_count = components[0].numpins
            bom_item.category = "ELECTRICAL"
            bom_item.charactistics.device_type = components[0].type
            bom_item.charactistics.category = "ELECTRICAL"
            bom_item.charactistics.component_class = "DISCRETE"
            if components[0].type == "Resistor":
                bom_item.charactistics.value = components[0].res_value
            elif components[0].type == "Capacitor":
                bom_item.charactistics.value = components[0].cap_value
            elif components[0].type == "Inductor":
                bom_item.charactistics.value = components[0].ind_value
            for cmp in components:
                bom_item.add_refdes(
                    component_name=cmp.refdes, placement_layer=cmp.placement_layer, package_def="", populate=True
                )
            self.bom.bom_items.append(bom_item)

    @pyaedt_function_handler()
    def add_layers_info(self):
        self.design_name = self._pedb._active_layout.GetCell().GetName()
        self.ecad.design_name = self.design_name
        self.ecad.cad_header.units = self.units
        self.ecad.cad_data.stackup.total_thickness = self.from_meter_to_units(
            self._pedb.stackup.get_layout_thickness(), self.units
        )
        self.ecad.cad_data.stackup.stackup_group.thickness = self.ecad.cad_data.stackup.total_thickness
        self.layers_name = list(self._pedb.stackup.signal_layers.keys())
        self.top_bottom_layers = [self.layers_name[0], self.layers_name[-1]]
        sequence = 0
        for layer_name in list(self._pedb.stackup.stackup_layers.keys()):
            sequence += 1
            self.content.add_layer_ref(layer_name)
            layer_color = self._pedb.stackup.layers[layer_name].color
            self.content.dict_colors.add_color(
                "{}".format(layer_name), str(layer_color[0]), str(layer_color[1]), str(layer_color[2])
            )
            # Ecad layers
            layer_type = "CONDUCTOR"
            conductivity = 5e6
            permitivity = 1
            loss_tg = 0
            embedded = "NOT_EMBEDDED"
            # try:
            material_name = self._pedb.stackup.layers[layer_name]._edb_layer.GetMaterial()
            edb_material = self._pedb.edb.Definition.MaterialDef.FindByName(self._pedb.db, material_name)
            material_type = "CONDUCTOR"
            if self._pedb.stackup.layers[layer_name].type == "dielectric":
                layer_type = "DIELPREG"
                material_type = "DIELECTRIC"
                permitivity = edb_material.GetProperty(self._pedb.edb.Definition.MaterialPropertyId.Permittivity)[
                    1
                ].ToDouble()
                loss_tg = edb_material.GetProperty(self._pedb.edb.Definition.MaterialPropertyId.DielectricLossTangent)[
                    1
                ].ToDouble()
                conductivity = 0
            if layer_type == "CONDUCTOR":
                conductivity = edb_material.GetProperty(self._pedb.edb.Definition.MaterialPropertyId.Conductivity)[
                    1
                ].ToDouble()
            self.ecad.cad_header.add_spec(
                name=layer_name,
                material=self._pedb.stackup.layers[layer_name]._edb_layer.GetMaterial(),
                layer_type=material_type,
                conductivity=str(conductivity),
                dielectric_constant=str(permitivity),
                loss_tg=str(loss_tg),
                embedded=embedded,
            )
            layer_position = "INTERNAL"
            if layer_name == self.top_bottom_layers[0]:
                layer_position = "TOP"
            if layer_name == self.top_bottom_layers[1]:
                layer_position = "BOTTOM"
            self.ecad.cad_data.add_layer(
                layer_name=layer_name, layer_function=layer_type, layer_side=layer_position, polarity="POSITIVE"
            )
            layer_thickness_with_units = self.from_meter_to_units(
                self._pedb.stackup.layers[layer_name].thickness, self.units
            )
            self.ecad.cad_data.stackup.stackup_group.add_stackup_layer(
                layer_name=layer_name, thickness=layer_thickness_with_units, sequence=str(sequence)
            )
            # except:
            #    pass
        self.ecad.cad_data.add_layer(layer_name="Drill", layer_function="DRILL", layer_side="ALL", polarity="POSITIVE")
        self.content.add_layer_ref("Drill")
        self.content.dict_colors.add_color("{}".format("Drill"), "255", "255", "255")

    @pyaedt_function_handler()
    def add_components(self):
        for item in self._pedb.core_components.components.values():
            self.ecad.cad_data.cad_data_step.add_component(item)

    @pyaedt_function_handler()
    def add_logical_nets(self):
        nets = [i for i in self._pedb.core_nets.nets.values()]
        for net in nets:
            self.ecad.cad_data.cad_data_step.add_logical_net(net)

    @pyaedt_function_handler()
    def add_layer_features(self):
        layers = {i: j for i, j in self._pedb.stackup.signal_layers.items()}
        padstack_instances = list(self._pedb.core_padstack.padstack_instances.values())
        padstack_defs = {i: k for i, k in self._pedb.core_padstack.padstacks.items()}
        polys = {i: j for i, j in self._pedb.core_primitives.primitives_by_layer.items()}
        for layer_name, layer in layers.items():
            self.ecad.cad_data.cad_data_step.add_layer_feature(layer, polys[layer_name])
        self.ecad.cad_data.cad_data_step.add_padstack_instances(padstack_instances, padstack_defs)

    @pyaedt_function_handler()
    def add_drills(self):
        via_list = [
            obj
            for obj in list(self._pedb.core_padstack.padstack_instances.values())
            if not obj.start_layer == obj.stop_layer
        ]
        l1 = len(list(self._pedb.stackup.signal_layers.keys()))

        self.ecad.cad_data.cad_data_step.add_drill_layer_feature(via_list, "DRILL_1-{}".format(l1))

    @pyaedt_function_handler()
    def from_meter_to_units(self, value, units):
        if isinstance(value, str):
            value = float(value)
        if isinstance(value, list):
            returned_list = []
            for val in value:
                if isinstance(val, str):
                    val = float(val)
                if units.lower() == "mm":
                    returned_list.append(round(val * 1000, 4))
                if units.lower() == "um":
                    returned_list.append(round(val * 1e6, 4))
                if units.lower() == "mils":
                    returned_list.append(round(val * 39370.079, 4))
                if units.lower() == "inch":
                    returned_list.append(round(val * 39.370079, 4))
                if units.lower() == "cm":
                    returned_list.append(round(val * 100, 4))
            return returned_list
        else:
            if units.lower() == "millimeter":
                return round(value * 1000, 4)
            if units.lower() == "micrometer":
                return round(value * 1e6, 4)
            if units.lower() == "mils":
                return round(value * 39370.079, 4)
            if units.lower() == "inch":
                return round(value * 39.370079, 4)
            if units.lower() == "centimeter":
                return round(value * 100, 4)

    @pyaedt_function_handler()
    def write_xml(self):
        if self.file_path:
            ipc = ET.Element("IPC-2581")
            ipc.set("revision", self.revision)
            ipc.set("xmlns", "http://webstds.ipc.org/2581")
            ipc.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
            ipc.set("xmlns:xsd", "http://www.w3.org/2001/XMLSchema")
            self.content.write_wml(ipc)
            # self.logistic_header.write_xml(ipc)
            # self.history_record.write_xml(ipc)
            # self.bom.write_xml(ipc)
            self.ecad.write_xml(ipc)
            try:
                ET.indent(ipc)
            except AttributeError:
                pass
            tree = ET.ElementTree(ipc)
            tree.write(self.file_path)
            return True if os.path.exists(self.file_path) else False



from pyaedt.generic.general_methods import ET


class StackupLayer(object):
    def __init__(self):
        self.layer_name = ""
        self.thickness = 0
        self.tol_minus = 0
        self.tol_plus = 0
        self.sequence = ""
        self.spec_ref = "SPEC_{}".format(self.layer_name)

    def write_xml(self, stackup_group):  # pragma no cover
        stackup_layer = ET.SubElement(stackup_group, "StackupLayer")
        stackup_layer.set("layerOrGroupRef", self.layer_name)
        stackup_layer.set("thickness", str(self.thickness))
        stackup_layer.set("tolPlus", str(self.tol_plus))
        stackup_layer.set("tolMinus", str(self.tol_minus))
        stackup_layer.set("sequence", self.sequence)
        spec_ref = ET.SubElement(stackup_layer, "SpecRef")
        spec_ref.set("id", self.layer_name)



from pyaedt.edb_core.ipc2581.ecad.cad_data.stackup_group import StackupGroup
from pyaedt.generic.general_methods import ET


class Stackup(object):
    def __init__(self):
        self.name = "PRIMARY"
        self.total_thickness = 0.0
        self.tol_plus = 0.0
        self.tol_min = 0.0
        self.where_measured = "METAL"
        self._stackup_group = StackupGroup()

    @property
    def stackup_group(self):
        return self._stackup_group

    @stackup_group.setter
    def stackup_group(self, value):  # pragma no cover
        if isinstance(value, list):
            if len([stack_grp for stack_grp in value if isinstance(stack_grp, StackupGroup)]):
                self._stackup_group = value

    def write_xml(self, cad_data):  # pragma no cover
        stackup = ET.SubElement(cad_data, "Stackup")
        stackup.set("name", self.name)
        stackup.set("overallThickness", str(self.total_thickness))
        stackup.set("tolPlus", str(self.tol_plus))
        stackup.set("tolMinus", str(self.tol_min))
        stackup.set("whereMeasured", self.where_measured)
        self.stackup_group.write_xml(stackup)



import re

from pyaedt.generic.general_methods import generate_unique_name
from pyaedt.generic.general_methods import pyaedt_function_handler


class Monitor:
    """Provides Icepak monitor methods."""

    def __init__(self, p_app):
        self._face_monitors = {}
        self._point_monitors = {}
        self._app = p_app
        self._omonitor = self._app.odesign.GetModule("Monitor")
        if self._app.design_properties:  # if is not a 3d comp/blank file
            aedtfile_monitor_dict = self._app.design_properties["Monitor"]["IcepakMonitors"].copy()
            del aedtfile_monitor_dict["NextUniqueID"]
            del aedtfile_monitor_dict["MoveBackwards"]
            if aedtfile_monitor_dict:
                self._load_monitor_objects(aedtfile_monitor_dict)

    @pyaedt_function_handler
    def _find_point(self, position):
        for point in self._app.oeditor.GetPoints():
            point_pos = self._app.oeditor.GetChildObject(point).GetPropValue("Position")
            if all(point_pos[2 * i + 1] == position[i] for i in range(3)):
                return point
        return None

    @pyaedt_function_handler
    def _generate_monitor_names(self, name, n):
        """Create list of names for monitor objects following Icepak naming rules.

        Parameters
        ----------
        name : str
            Name for the first monitor object
        n : int
            Number of names to be generated

        Returns
        -------
        list
            List of names
        """
        if n == 1 and name not in self.all_monitors:
            return [name]
        else:
            if name not in self.all_monitors:
                names = [name]
            else:
                names = []
                n += 1
            j = 1
            if re.search(r"\d+$", name) is not None:
                j = int(re.search(r"\d+$", name).group(0)) + 1
            n_names_left = n - 1
            while n_names_left:
                i = n - 1 - n_names_left
                candidate_name = re.sub(r"\d+$", "", name) + str(i + j)
                if candidate_name not in self.all_monitors:
                    names.append(candidate_name)
                    n_names_left -= 1
                n += 1
            return names

    @pyaedt_function_handler
    def _load_monitor_objects(self, aedtfile_monitor_dict):
        quantities_dict = {  # pragma: no cover
            8: "Speed",
            9: "Pressure",
            10: "TKE",
            11: "Epsilon",
            12: "ViscosityRatio",
            16: "MassFlow",
            17: "VolumeFlow",
            18: "WallYPlus",
            19: "Temperature",
            20: "K_X",
            21: "K_Y",
            22: "K_Z",
            29: "HeatFlux",
            31: "HeatFlowRate",
        }
        for monitor_name, monitor_prop in aedtfile_monitor_dict.items():
            if "Faces" in monitor_prop.keys():
                self._face_monitors[monitor_name] = FaceMonitor(
                    monitor_name,
                    "Face",
                    monitor_prop["Faces"][0],
                    [quantities_dict[i] for i in monitor_prop["Quantities"]],
                    self._app,
                )
            elif "Objects" in monitor_prop.keys() and monitor_prop["Type"] == 2:
                self._point_monitors[monitor_name] = PointMonitor(
                    monitor_name,
                    "Object",
                    self._app.oeditor.GetObjectNameByID(int(monitor_prop["Objects"][0])),
                    [quantities_dict[i] for i in monitor_prop["Quantities"]],
                    self._app,
                )
            elif "Objects" in monitor_prop.keys():
                self._face_monitors[monitor_name] = FaceMonitor(
                    monitor_name,
                    "Surface",
                    self._app.oeditor.GetObjectNameByID(int(monitor_prop["Objects"][0])),
                    [quantities_dict[i] for i in monitor_prop["Quantities"]],
                    self._app,
                )
            elif "Points" in monitor_prop.keys():
                point_name = self._find_point(
                    self._app.odesign.GetChildObject("Monitor")
                    .GetChildObject(monitor_name)
                    .GetPropValue("Location")
                    .split(", ")
                )
                self._point_monitors[monitor_name] = PointMonitor(
                    monitor_name,
                    "Point",
                    point_name,
                    [quantities_dict[i] for i in monitor_prop["Quantities"]],
                    self._app,
                )
            else:
                self._point_monitors[monitor_name] = PointMonitor(
                    monitor_name,
                    "Vertex",
                    monitor_prop["Vertices"][0],
                    [quantities_dict[i] for i in monitor_prop["Quantities"]],
                    self._app,
                )
        return True

    @pyaedt_function_handler
    def get_icepak_monitor_object(self, monitor_name):
        """Get Icepak monitor object.

        Returns
        -------
        oEditor COM Object
        """
        return self._app.odesign.GetChildObject("Monitor").GetChildObject(monitor_name)

    @property
    def face_monitors(self):
        """Get face monitor objects.

        Returns
        -------
        dict
            Face monitor objects dictionary.

        """
        return self._face_monitors

    @property
    def point_monitors(self):
        """Get face monitor objects.

        Returns
        -------
        dict
            Point monitor objects dictionary.

        """
        return self._point_monitors

    @property
    def all_monitors(self):
        """Get all monitor objects.

        Returns
        -------
        dict
            Monitor objects dictionary.

        """
        out_dict = {}
        out_dict.update(self.face_monitors)
        out_dict.update(self.point_monitors)
        return out_dict

    @pyaedt_function_handler()
    def assign_point_monitor(self, point_position, monitor_quantity="Temperature", monitor_name=None):
        """Create and assign a point monitor.

        Parameters
        ----------
        point_position : list
            List of the ``[x, y, z]`` coordinates for the point or list of list of coordinates.
        monitor_quantity : str or list, optional
            Quantity being monitored.  The default is ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor. The default is ``None``, in which case
            the name is randomly generated.

        Returns
        -------
        str or list
            Monitor name or list of monitor names when successful, ``False`` when failed.

        References
        ----------
        >>> oModule.AssignPointMonitor

        Examples
        --------
        Create two temperature monitor at the points ``[0, 0, 0]`` and ``[1, 1, 1]``.

        >>> icepak.monitor.assign_point_monitor([[0,0,0], [1, 1, 1]], monitor_name="monitor1")
        ['monitor1', 'monitor2']

        """
        point_names = []
        if not isinstance(point_position[0], list):
            point_position = [point_position]
        if not isinstance(monitor_quantity, list):
            monitor_quantity = [monitor_quantity]
        for p_p in point_position:
            point_name = generate_unique_name("Point")
            self._app.modeler.oeditor.CreatePoint(
                [
                    "NAME:PointParameters",
                    "PointX:=",
                    self._app.modeler._arg_with_dim(p_p[0]),
                    "PointY:=",
                    self._app.modeler._arg_with_dim(p_p[1]),
                    "PointZ:=",
                    self._app.modeler._arg_with_dim(p_p[2]),
                ],
                ["NAME:Attributes", "Name:=", point_name, "Color:=", "(143 175 143)"],
            )
            point_names.append(point_name)
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        self._omonitor.AssignPointMonitor(
            ["NAME:" + monitor_name, "Quantities:=", monitor_quantity, "Points:=", point_names]
        )
        try:
            monitor_names = self._generate_monitor_names(monitor_name, len(point_position))
            for i, mn in enumerate(monitor_names):
                self._point_monitors[mn] = PointMonitor(mn, "Point", point_names[i], monitor_quantity, self._app)
        except:  # pragma: no cover
            return False
        if len(monitor_names) == 1:
            return monitor_names[0]
        else:
            return monitor_names

    @pyaedt_function_handler()
    def assign_point_monitor_to_vertex(self, vertex_id, monitor_quantity="Temperature", monitor_name=None):
        """Create and assign a point monitor to a vertex.

        Parameters
        ----------
        vertex_id : int or list
            ID of the vertex or list of IDs.
        monitor_quantity : str or list, optional
            Quantity being monitored.  The default is ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor. The default is ``None``, in which case
            the name is randomly generated.

        Returns
        -------
        str or list
            Monitor name or list of monitor names when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.AssignPointMonitor

        """
        if isinstance(vertex_id, int):
            vertex_id = [vertex_id]
        if isinstance(monitor_quantity, str):
            monitor_quantity = [monitor_quantity]
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        self._omonitor.AssignPointMonitor(
            ["NAME:" + monitor_name, "Quantities:=", monitor_quantity, "Vertices:=", vertex_id]
        )
        try:
            monitor_names = self._generate_monitor_names(monitor_name, len(vertex_id))
            for i, mn in enumerate(monitor_names):
                self._point_monitors[mn] = PointMonitor(mn, "Vertex", vertex_id[i], monitor_quantity, self._app)
        except:  # pragma: no cover
            return False
        if len(monitor_names) == 1:
            return monitor_names[0]
        else:
            return monitor_names

    @pyaedt_function_handler()
    def assign_surface_monitor(self, surface_name, monitor_quantity="Temperature", monitor_name=None):
        """Assign a surface monitor.

        Parameters
        ----------
        surface_name : str or list
            Name of the surface or list of names.
        monitor_quantity : str or list, optional
            Quantity being monitored.  The default is ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor. The default is ``None``, in which case
            the name is randomly generated.

        Returns
        -------
        str or list
            Monitor name or list of monitor names when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.AssignFaceMonitor

        Examples
        --------

        Create a rectangle named ``"Surface1"`` and assign a temperature monitor to that surface.

        >>> surface = icepak.modeler.create_rectangle(icepak.PLANE.XY,
        ...                                           [0, 0, 0], [10, 20], name="Surface1")
        >>> icepak.assign_surface_monitor("Surface1", monitor_name="monitor")
        'monitor'
        """
        if isinstance(surface_name, str):
            surface_name = [surface_name]
        if isinstance(monitor_quantity, str):
            monitor_quantity = [monitor_quantity]
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        self._omonitor.AssignFaceMonitor(
            ["NAME:" + monitor_name, "Quantities:=", monitor_quantity, "Objects:=", surface_name]
        )
        try:
            monitor_names = self._generate_monitor_names(monitor_name, len(surface_name))
            for i, mn in enumerate(monitor_names):
                self._face_monitors[mn] = FaceMonitor(mn, "Surface", surface_name[i], monitor_quantity, self._app)
        except:  # pragma: no cover
            return False
        if len(monitor_names) == 1:
            return monitor_names[0]
        else:
            return monitor_names

    @pyaedt_function_handler()
    def assign_face_monitor(self, face_id, monitor_quantity="Temperature", monitor_name=None):
        """Assign a face monitor.

        Parameters
        ----------
        face_id : int or list
            Face id or list of ids
        monitor_quantity : str or list, optional
            Quantity being monitored.  The default is ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor. The default is ``None``, in which case
            the name is randomly generated.

        Returns
        -------
        str or list
            Monitor name or list of monitor names when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.AssignFaceMonitor
        """
        if isinstance(face_id, int):
            face_id = [face_id]
        if isinstance(monitor_quantity, str):
            monitor_quantity = [monitor_quantity]
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        self._omonitor.AssignFaceMonitor(["NAME:" + monitor_name, "Quantities:=", monitor_quantity, "Faces:=", face_id])
        try:
            monitor_names = self._generate_monitor_names(monitor_name, len(face_id))
            for i, mn in enumerate(monitor_names):
                self._face_monitors[mn] = FaceMonitor(mn, "Face", face_id[i], monitor_quantity, self._app)
        except:  # pragma: no cover
            return False
        if len(monitor_names) == 1:
            return monitor_names[0]
        else:
            return monitor_names

    @pyaedt_function_handler()
    def assign_point_monitor_in_object(self, name, monitor_quantity="Temperature", monitor_name=None):
        """Assign a point monitor in the centroid of a specific object.

        Parameters
        ----------
        name : str or list
            Name of the object to assign monitor point to.
        monitor_quantity : str or list, optional
            Quantity being monitored.  The default is ``"Temperature"``.
        monitor_name : str, optional
            Name of the monitor. The default is ``None``, in which case
            the name is randomly generated.

        Returns
        -------
        str or list
            Monitor name or list of names when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.AssignPointMonitor

        Examples
        --------

        Create a box named ``"BlockBox1"`` and assign a temperature monitor point to that object.

        >>> box = icepak.modeler.create_box([1, 1, 1], [3, 3, 3], "BlockBox1", "copper")
        >>> icepak.assign_point_monitor(box.name, monitor_name="monitor2")
        "'monitor2'
        """
        if not isinstance(name, list):
            name = [name]
        if not isinstance(monitor_quantity, list):
            monitor_quantity = [monitor_quantity]
        name_sel = self._app.modeler.convert_to_selections(name, True)
        original_monitors = list(self._app.odesign.GetChildObject("Monitor").GetChildNames())
        if not monitor_name:
            monitor_name = generate_unique_name("Monitor")
        elif monitor_name in original_monitors:
            monitor_name = generate_unique_name(monitor_name)
        existing_names = list(set(name_sel).intersection(self._app.modeler.object_names))
        if existing_names:
            self._omonitor.AssignPointMonitor(
                ["NAME:" + monitor_name, "Quantities:=", monitor_quantity, "Objects:=", existing_names]
            )
        else:
            self._app.logger.error("Object is not present in the design")
            return False
        try:
            monitor_names = self._generate_monitor_names(monitor_name, len(existing_names))
            for i, mn in enumerate(monitor_names):
                self._point_monitors[mn] = PointMonitor(mn, "Object", existing_names[i], monitor_quantity, self._app)
        except:  # pragma: no cover
            return False
        if len(monitor_names) == 1:
            return monitor_names[0]
        else:
            return monitor_names

    @pyaedt_function_handler()
    def delete_monitor(self, monitor_name):
        """Delete monitor object.

        Parameters
        ----------
        monitor_name : str
            Name of the monitor object to delete

        Returns
        -------
        bool
            ``False`` if succesflul, else ``False``.

        References
        ----------

        >>> oModule.DeleteMonitors

        """
        try:
            self._omonitor.DeleteMonitors([monitor_name])
            try:
                del self._face_monitors[monitor_name]
            except KeyError:
                del self._point_monitors[monitor_name]
            return True
        except:
            return False

    @pyaedt_function_handler()
    def get_monitor_object_assignment(self, monitor):
        """
        Get the object that the monitor is applied to.

        Parameters
        ----------
        monitor : str or FaceMonitor or PointMonitor object
           Monitor object or monitor object name.

        Returns
        -------
        str
            Name of the object.
        """
        if isinstance(monitor, str):
            monitor = self.all_monitors[monitor]
        m_type = monitor.type
        if m_type == "Face":
            return self._app.oeditor.GetObjectNameByFaceID(monitor.id)
        elif m_type == "Vertex":
            return self._app.oeditor.GetObjectNameByVertexID(monitor.id)
        else:
            return monitor.id

    def _delete_removed_monitors(self):
        existing_monitors = self._app.odesign.GetChildObject("Monitor").GetChildNames()
        for j in [self._face_monitors, self._point_monitors]:
            for i in list(j):
                if i not in existing_monitors:
                    del j[i]
                    self._app.logger.info(
                        i + " monitor object lost its assignment due to geometry modifications and has been deleted."
                    )

    @pyaedt_function_handler()
    def insert_monitor_object_from_dict(self, monitor_dict, mode=0):
        """Insert a monitor.

        Parameters
        ----------
        monitor_dict : dict
           Dictionary containing monitor object information.
        mode : int
            Integer to select the information to handle. To identify the faces, vertices,
            surfaces, and object to which to assign the monitor to, you can use:
            - ids and names, mode=0, required dict keys: "Name", "Type", "ID", "Quantity".
            - positions, mode=1, required dict keys: "Name", "Type", "Geometry Assignment", "Location", "Quantity".

        Returns
        -------
        str
            Name of the monitor object.
        """
        m_case = monitor_dict["Type"]
        m_quantity = monitor_dict["Quantity"]
        m_name = monitor_dict["Name"]
        m_object = None
        if mode == 1:
            if monitor_dict["Geometry Assignment"] not in self._app.modeler.object_names:
                return False
            if m_case == "Face":
                for f in self._app.modeler.get_object_from_name(monitor_dict["Geometry Assignment"]).faces:
                    if f.center == monitor_dict["Location"]:
                        m_object = f.id
                        break
            elif m_case == "Vertex":
                for v in self._app.modeler.get_object_from_name(monitor_dict["Geometry Assignment"]).vertices:
                    if v.position == monitor_dict["Location"]:
                        m_object = v.id
                        break
            elif m_case == "Object":
                m_object = monitor_dict["Geometry Assignment"]
            elif m_case == "Surface":
                m_object = monitor_dict["Geometry Assignment"]
        elif mode == 0:
            m_object = monitor_dict["ID"]
        else:
            self._app.logger.error("Only modes supported are 0 and 1")
        if m_object is None:  # pragma: no cover
            self._app.logger.error("{} monitor object could not be restored".format(m_name))
            return False
        self._app.configurations.update_monitor(m_case, m_object, m_quantity, m_name)
        self._app.logger.info("{} monitor object restored".format(m_name))
        return m_name


class ObjectMonitor:
    """Provides Icepak Monitor methods and properties."""

    def __init__(self, monitor_name, monitor_type, monitor_id, quantity, app):
        self._name = monitor_name
        self._type = monitor_type
        self._id = monitor_id
        self._quantities = quantity
        self._app = app

    @property
    def geometry_assignment(self):
        """
        Get the geometry assignment for the monitor object.

        Returns
        -------
        str
        """
        return self._app.monitor.get_monitor_object_assignment(self)

    @property
    def name(self):
        """
        Get the name of the monitor object.

        Returns
        -------
        str
        """
        return self._name

    @property
    def id(self):
        """
        Get the name, or id of geometry assignment.

        Returns
        -------
        str or int
        """
        return self._id

    @property
    def properties(self):
        """
        Get a dictionary of properties.

        Returns
        -------
        dict
        """
        return {
            "Name": self.name,
            "Object": self._app.odesign.GetChildObject("Monitor").GetChildObject(self.name),
            "Type": self.type,
            "ID": self.id,
            "Location": self.location,
            "Quantity": self.quantities,
            "Geometry Assignment": self.geometry_assignment,
        }

    @pyaedt_function_handler
    def delete(self):
        """
        Delete a monitor object.

        Returns
        -------
        bool
            ``True`` if successful.
        """
        self._app.monitor.delete_monitor(self.name)
        return True

    @property
    def quantities(self):
        """
        Get the quantities being monitored.

        Returns
        -------
        list
        """
        return self._quantities

    @property
    def type(self):
        """
        Get the monitor type.

        Returns
        -------
        str
        """
        return self._type


class PointMonitor(ObjectMonitor):
    """Provides Icepak point monitor methods and properties."""

    def __init__(self, monitor_name, monitor_type, point_id, quantity, app):
        ObjectMonitor.__init__(self, monitor_name, monitor_type, point_id, quantity, app)

    @property
    def location(self):
        """
        Get the monitor point location.

        Returns
        -------
        list
            List of floats containing [x, y, z] position.
        """
        return [
            float(i.strip(self._app.modeler.model_units))
            for i in self._app.odesign.GetChildObject("Monitor")
            .GetChildObject(self._name)
            .GetPropValue("Location")
            .split(", ")
        ]


class FaceMonitor(ObjectMonitor):
    """Provides Icepak face monitor properties and methods."""

    def __init__(self, monitor_name, monitor_type, face_id, quantity, app):
        ObjectMonitor.__init__(self, monitor_name, monitor_type, face_id, quantity, app)

    @property
    def location(self):
        """
        Get the monitor location in terms of face or surface center.

        Returns
        -------
        list
            List of floats containing [x, y, z] position.
        """
        if self.type == "Face":
            for f in self._app.modeler.get_object_from_name(self.geometry_assignment).faces:
                if f.id == self.id:
                    return f.center
        elif self.type == "Surface":
            return self._app.modeler.get_object_from_name(self.geometry_assignment).faces[0].center
